import json
import codecs
import jinja2
from os import path
from pathlib import Path
from copy import deepcopy
from packaging import version
from lyrebird import log as nlog
from lyrebird import application

from lyrebird.utils import RedisDict
from lyrebird.config.diff_mode import SettingDiffMode
from lyrebird.config.checker_switch import SettingCheckerSwitch

from .keywords import *

logger = nlog.get_logger()

config_template = {
    "version": "0.10.5",
    "proxy.filters": [],
    "proxy.port": 4272,
    "mock.port": 9090,
    "mock.data": path.join("{{current_dir}}", "mock_data", "personal"),
    "mock.proxy_headers": {
        "scheme": "MKScheme",
        "host": "MKOriginHost",
        "port": "MKOriginPort"
    }
}

CONFIG_FUNC_MAP = {
    'checker.switch': SettingCheckerSwitch,
    'mock.mode': SettingDiffMode
}

personal_config_template = {
    "event.broken_database_path_list": []
}


class ConfigManager():
    ROOT = Path('~/.lyrebird').expanduser()
    DEFAULT_FILENAME = 'conf.json'
    DEFAULT_PERSONAL_FILENAME = 'personal_conf.json'
    BASE_CONFIG = ROOT / DEFAULT_FILENAME
    PERSONAL_CONFIG = ROOT / DEFAULT_PERSONAL_FILENAME
    FORBIDDEN_MODIFY_FIELDS_IN_CONFIG = set(['version', 'proxy.port', 'mock.port'])

    def __init__(self, conf_path_list=None, custom_conf=None):
        self.config = config_template
        self.config_root = self.ROOT
        self.conf_file = self.BASE_CONFIG
        self.config_list = []
        # Current personal config only supports checking whether lyrebird.db is broken.
        self.personal_config = personal_config_template
        self.personal_conf_file = self.PERSONAL_CONFIG

        self.update_base_config()
        self.read_config()
        if conf_path_list:
            for conf_path in conf_path_list:
                self.update_conf_source(conf_path)
        if custom_conf:
            self.update_conf_custom(custom_conf)

        if self.config.get('enable_multiprocess', False):
            try:
                self.config = RedisDict(data=self.config,
                                        host=self.config.get('redis_host', '127.0.0.1'),
                                        port=self.config.get('redis_port', 6379),
                                        db=self.config.get('redis_db', 0))
            except Exception as e:
                self.config['enable_multiprocess'] = False
                logger.error(f'Start enable multiprocess failed, Redis connection error:\n{e}')

        self.initialize_personal_config()

    def update_conf_source(self, path):
        input_path: Path = Path(path).expanduser().absolute()
        if input_path.is_dir():
            input_root = input_path
            input_file = input_path / self.DEFAULT_FILENAME
        else:
            input_root = input_path.parent
            input_file = input_path

        if not input_file.exists():
            logger.error(f'Config {input_file} not found!')
        else:
            self.config_root = input_root
            self.conf_file = input_file
            self.read_config()

    def update_conf_custom(self, custom_conf):
        self.add_config(custom_conf, rank=-1)

    def contains_forbidden_modify_field(self, update_conf: dict):
        union_fields = self.FORBIDDEN_MODIFY_FIELDS_IN_CONFIG & update_conf.keys()
        return union_fields if len(union_fields) > 0 else None

    def override_config_field(self, update_conf: dict):
        """
        Update Application config by ``config.update(update_conf)``.
        If update_conf contains ``FORBIDDEN_MODIFY_FIELDS_IN_CONFIG``, raise ``ConfigException``.
        """
        if not update_conf:
            return

        forbidden_modify_fields = self.contains_forbidden_modify_field(update_conf)
        if forbidden_modify_fields:
            raise ConfigException(f'Config field cannot be modified: {forbidden_modify_fields}')

        update_level = self.config.get('config.update_config.level', 1)

        logger.debug(f'Need update config fields: {update_conf}')
        self.add_config(update_conf, type='api_patch', level=update_level, apply_now=True)
        logger.debug(f'Update done. config: {self.config}')

        application.server['event'].publish('config_update', {'config_update' : {'data': str(update_conf)}})

    def read_config(self):
        template_env = jinja2.Environment(loader=jinja2.FileSystemLoader(str(self.config_root)))
        template = template_env.get_template(self.conf_file.name)
        current_dir = str(self.config_root)
        download_dir = str(self.ROOT / 'downloads')
        conf_str = template.render(current_dir=json.dumps(current_dir).strip('"'), download_dir=json.dumps(download_dir).strip('"'))
        loaded_config = json.loads(conf_str)

        self.add_config(loaded_config, rank=-10)

    def write_config(self):
        self.config_root.mkdir(parents=True, exist_ok=True)
        with codecs.open(self.conf_file, 'w', 'utf-8') as f:
            f.write(json.dumps(self.config, indent=4, ensure_ascii=False))

    def update_base_config(self):
        if self.BASE_CONFIG.exists() and self.BASE_CONFIG.is_file():
            with codecs.open(self.BASE_CONFIG, 'r', 'utf-8') as f:
                base_conf = json.load(f)
                if version.parse(base_conf.get('version', '0.0.0')) < version.parse(config_template.get('version', '0.0.0')):
                    self.write_config()
        else:
            self.write_config()

    def add_config(self, config_dict: dict, rank=0, type='', override_same_type=False, level=1, apply_now=False) -> None:
        config = Config(config_dict, level, self.config)
        config.type = type

        if override_same_type:
            for c in self.config_list[::-1]:
                if c.type == type:
                    self.config_list.remove(c)
        self.config_list.append(config)

        self.merge_config(self.config, config.config, level=level, apply_now=apply_now)

    def merge_config(self, origin_config, new_config, level=-1, apply_now=False):
        for key_child in list(new_config.keys()):
            self.merge_generator(key_child, origin_config, new_config, level)

        if apply_now:
            self.add_each_config_item(new_config)

    def merge_generator(self, key, origin_config, new_config, level):
        if level == 0:
            return

        if key not in origin_config:
            origin_config[key] = deepcopy(new_config[key])

        elif origin_config[key] == new_config[key]:
            return

        elif level == 1:
            origin_config[key] = deepcopy(new_config[key])

        elif type(origin_config[key]) != type(new_config[key]):
            origin_config[key] = deepcopy(new_config[key])

        elif isinstance(new_config[key], list):
            origin_config[key] = deepcopy(new_config[key])

        elif isinstance(new_config[key], dict):
            for key_child in list(new_config[key].keys()):
                self.merge_generator(key_child, origin_config[key], new_config[key], level-1)

        else:
            origin_config[key] = new_config[key]

    def add_each_config_item(self, config):
        # 处理第一层的key
        for key, value in config.items():
            if key not in CONFIG_FUNC_MAP:
                continue
            CONFIG_FUNC_MAP[key].add(value)

    def remove_config(self, config=None, type='', apply_now=False):
        remove_config = None
        for c in self.config_list[::-1]:
            if c.type == type and (config is None or config == c.config):
                remove_config = c
                break

        if remove_config is None:
            logger.error("No matching config found in config_list.")
            return

        self.unmerge_config(self.config, remove_config.config, level=remove_config.level, apply_now=apply_now)

        self.merge_config(self.config, remove_config.previous_config, level=remove_config.level, apply_now=apply_now)

        # todo handle the config added after remove_config

        self.config_list.remove(remove_config)

    def unmerge_config(self, origin_config, remove_config, level=-1, apply_now=False):
        for key_child in list(remove_config.keys()):
            self.unmerge_generator(key_child, origin_config, remove_config, level=level)

        if apply_now:
            self.remove_each_config_item(remove_config)

    def unmerge_generator(self, key, origin_config, remove_config, level):
        if level == 0:
            return

        if key not in origin_config:
            return

        if level == 1:
            origin_config.pop(key, None)

        elif isinstance(remove_config[key], dict):
            for key_child in list(remove_config[key].keys()):
                self.unmerge_generator(key_child, origin_config[key], remove_config[key], level-1)

        else:
            origin_config.pop(key, None)

    def remove_each_config_item(self, config):
        # 处理第一层的key
        for key, value in config.items():
            if key not in CONFIG_FUNC_MAP:
                continue
            CONFIG_FUNC_MAP[key].remove(value)
    
    def initialize_personal_config(self):
        if not self.personal_conf_file.exists():
            self.write_personal_config()
        self.personal_config = self.read_personal_config()
    
    def update_personal_config(self, config_dict: dict):
        self.personal_config.update(config_dict)
        self.write_personal_config()

    def read_personal_config(self):
        try:
            with codecs.open(self.personal_conf_file, 'r', 'utf-8') as f:
                return json.load(f)
        except json.JSONDecodeError:
            logger.error(f'Failed to parse personal config file {self.personal_conf_file}!')
            return personal_config_template

    def write_personal_config(self):
        try:
            with codecs.open(self.personal_conf_file, 'w', 'utf-8') as f:
                f.write(json.dumps(self.personal_config, indent=4, ensure_ascii=False))
        except Exception as e:
            logger.error(f'Failed to write personal config: {e}')


class Config:
    def __init__(self, config, level=1, current_config={}):
        self.rank = 0
        self.type = ''
        self.config = config
        self.level = level
        self.previous_config = self._get_previuos_config(current_config, level)

    def _get_previuos_config(self, current_config, level):
        if not current_config:
            return {}
        previous_config = {}
        for key in self.config.keys():
            self._get_previous_config_generator(key, self.config, current_config, previous_config, level)
        return previous_config
    
    def _get_previous_config_generator(self, key, config, current_config, previous_config, level):
        if not current_config or key not in current_config:
           return

        if level != 1 and isinstance(config[key], dict):
            previous_config[key] = {}
            for key_child in config[key].keys():
                self._get_previous_config_generator(key_child, config[key], current_config.get(key), previous_config[key], level-1)
        else:
            previous_config[key] = deepcopy(current_config.get(key))


class ConfigException(Exception):
    pass



import json
from flask_restful import Resource
from flask import jsonify, request, Response, stream_with_context
from lyrebird import application
from lyrebird.db.event_converter import export_from_event

# Default event page size
PAGE_SIZE = 20


class Event(Resource):

    def get(self, channel=None, page=0, event_id=None, search_str=''):
        if request.args.get('q'):
            search_str = request.args.get('q')
        db = application.server['db']
        channel_rules = []
        if channel:
            channel_rules = channel.split('+')
        if event_id:
            page = db.get_page_index_by_event_id(event_id, channel_rules, limit=PAGE_SIZE)
        events = db.get_event(channel_rules, offset=page*PAGE_SIZE, limit=PAGE_SIZE, search_str=search_str)
        page_count = db.get_page_count(channel_rules, page_size=PAGE_SIZE, search_str=search_str)
        result = []
        for event in events:
            event_str = event.json()

            # Import decoder for decoding the requested content
            if event_str.get('channel') == 'flow':
                content = json.loads(event_str['content'])
                application.encoders_decoders.decoder_handler(content['flow'])
                event_str['content'] = json.dumps(content, ensure_ascii=False)

            result.append(event_str)
        return application.make_ok_response(events=result, page=page, page_count=page_count, page_size=PAGE_SIZE, channel=channel_rules)

    def post(self, channel):
        message = request.json
        event_manager = application.server['event']
        event_manager.publish(channel, message)
        return application.make_ok_response()

    def delete(self):
        db_server = application.server['db']
        db_server.reset()
        return application.make_ok_response()


class Channel(Resource):

    def get(self, mode=None):
        if not mode:
            db = application.server['db']
            channel_list = db.get_channel_list()
            return jsonify([item[0] for item in channel_list])

        elif mode == 'default':
            channel = application.config.get('event.default_channel', [])
            selected_channel = application.config.get('event.selected_channel', [])
            return application.make_ok_response(data=channel, selected=selected_channel)

    def post(self):
        filters = request.json.get('filters')
        application._cm.config['event.selected_channel'] = filters


class EventExport(Resource):
    
    def get(self, event_id):
        # TODO: export event by event_id
        pass

    def post(self, event_id=None):
        if not request.json.get('export'):
            return application.make_fail_response('Missing required argument: export')
        try:
            filename, output_gen = export_from_event(request.json)
        except Exception as e:
            return application.make_fail_response('Convert data to stream error: {e}')
        
        res = Response(stream_with_context(output_gen()), mimetype="application/octet-stream")
        res.headers['Content-Disposition'] = f'attachment; filename={filename}'
        return res


class EventFileInfo(Resource):

    def get(self):
        db = application.server['db']
        file_info = dict()
        if db is not None:
            file_info = db.get_database_info()
        return application.make_ok_response(file_info=file_info)



from .gzip import GzipHandler
from .default import DefaultHandler
from lyrebird.log import get_logger

logger = get_logger()

content_encoding_handlers = {
    'gzip': GzipHandler,
}

def _get_matched_action(content_encoding_name):
    if content_encoding_name in content_encoding_handlers:
        return content_encoding_handlers[content_encoding_name]
    return DefaultHandler

def origin2flow(content_encoding, request_data, chain=None):
    func = _get_matched_action(content_encoding)
    try:
        _data = func.origin2flow(request_data)
    except Exception as e:
        func = DefaultHandler
        _data = func.origin2flow(request_data)
        logger.warning(f'Convert Content-Encoding: {content_encoding} data origin2flow failed! {e}')
    finally:
        if chain and isinstance(chain, list):
            chain.append(func)

    return _data

def flow2origin(content_encoding, flow_data):
    func = _get_matched_action(content_encoding)
    try:
        _data = func.flow2origin(flow_data)
    except Exception as e:
        _data = DefaultHandler.flow2origin(flow_data)
        logger.warning(f'Convert Content-Encoding: {content_encoding} data flow2origin failed! {e}')
    return _data

def origin2string(content_encoding, request_data):
    func = _get_matched_action(content_encoding)
    try:
        _data = func.origin2string(request_data)
    except Exception as e:
        func = DefaultHandler
        _data = func.origin2string(request_data)
        logger.warning(f'Convert Content-Encoding: {content_encoding} data origin2string failed! {e}')

    return _data



from hashlib import md5
from lyrebird import application

class SettingsTemplate:

    def __init__(self):
        self.display = True
        self.name = ''
        self.setting_items = []
        self.title = ''
        self.notice = ''
        self.submit_text = ''
        self.language = ''
        self.category = ''
        self.category_md5 = md5(self.category.encode(encoding='UTF-8')).hexdigest()
        self.manager = application.server['settings']

    def getter(self):
        pass

    def setter(self, data):
        pass

    def restore(self):
        pass

    def destory(self):
        pass

    def load_finished(self):
        pass

    def load_prepared(self):
        pass

class SettingItemTemplate:
    def __init__(self):
        self.name = ''
        self.title = ''
        self.subtitle = ''
        # enum, value should in {'list', 'dict', 'text', 'selector'}
        self.category = ''
        # only used in selector
        self.options = []
    
    def get_data(self):
        template_res = {
                        'name': self.name,
                        'title': self.title,
                        'subtitle': self.subtitle,
                        'category': self.category,
                        'data': '',
                        'options': self.options
                    }
        return template_res

    def set_data(self):
        return True



from lyrebird import application
from lyrebird.log import get_logger
from pathlib import Path
from copy import deepcopy
from importlib import machinery
import traceback
import datetime
import signal
from lyrebird.base_server import ProcessServer
from concurrent.futures import ThreadPoolExecutor
from lyrebird import application
from lyrebird.compatibility import prepare_application_for_monkey_patch, monkey_patch_application


logger = get_logger()

class Reporter(ProcessServer):

    def __init__(self):
        super().__init__()
        self.scripts = []
        self.workspace = application.config.get('reporter.workspace')
        self.report_queue = application.sync_manager.get_multiprocessing_queue()
        if not self.workspace:
            logger.debug(f'reporter.workspace not set.')
        elif not application.config.get('enable_multiprocess', False):
            self.scripts = self._read_reporter(self.workspace)
            logger.debug(f'Load statistics scripts {self.scripts}')

    def _read_reporter(self, workspace):
        target_dir = Path(workspace)
        scripts = []
        if not target_dir.exists():
            logger.error('Reporter workspace not found')
        for report_script_file in target_dir.iterdir():
            if report_script_file.name.startswith('_'):
                continue
            if not report_script_file.is_file():
                logger.warning(f'Skip report script: is not a file, {report_script_file}')
                continue
            if report_script_file.suffix != '.py':
                logger.warning(f'Skip report script: is not a python file, {report_script_file}')
                continue
            try:
                loader = machinery.SourceFileLoader('reporter_script', str(report_script_file))
                _script_module = loader.load_module()
            except Exception:
                logger.warning(
                    f'Skip report script: load script failed, {report_script_file}\n{traceback.format_exc()}')
                continue
            if not hasattr(_script_module, 'report'):
                logger.warning(f'Skip report script: not found a report method in script, {report_script_file}')
                continue
            if not callable(_script_module.report):
                logger.warning(f'Skip report script: report method not callable, {report_script_file}')
                continue
            scripts.append(_script_module.report)
        return scripts
    
    def start(self):
        if not application.config.get('enable_multiprocess', False):
            return
        self.process_namespace = prepare_application_for_monkey_patch()
        self.async_obj['report_queue'] = self.report_queue
        self.async_obj['workspace'] = self.workspace
        self.async_obj['process_namespace'] = self.process_namespace
        super().start()

    def run(self, async_obj, config, *args, **kwargs):

        signal.signal(signal.SIGINT, signal.SIG_IGN)

        workspace = async_obj['workspace']
        reportor_queue = async_obj['report_queue']

        monkey_patch_application(async_obj)
        scripts = self._read_reporter(workspace)

        self.thread_executor = ThreadPoolExecutor(max_workers=10)

        self.running = True

        while self.running:
            try:
                data = reportor_queue.get()
                if not data:
                    break
                new_data = deepcopy(data)
                for script in scripts:
                    try:
                        self.thread_executor.submit(script, new_data)
                    except Exception:
                        print(f'Send report failed:\n{traceback.format_exc()}')
            except Exception:
                logger.error(f'Reporter run error:\n{traceback.format_exc()}')

    def report(self, data):
        if self.running:
            self.report_queue.put(data)
        else:
            task_manager = application.server.get('task')

            def send_report():
                new_data = deepcopy(data)
                for script in self.scripts:
                    try:
                        script(new_data)
                    except Exception:
                        logger.error(f'Send report failed:\n{traceback.format_exc()}')
            task_manager.add_task('send-report', send_report)


def _page_out():

    if hasattr(application.sync_namespace,'last_page') and hasattr(application.sync_namespace,'last_page_in_time'):
        duration = (datetime.datetime.now() - application.sync_namespace.last_page_in_time).total_seconds()
        application.server['event'].publish('system', {
            'system': {
                'action': 'page.out', 'page': application.sync_namespace.last_page, 'duration': duration
            }
        })

        # TODO remove below
        application.reporter.report({
            'action': 'page.out',
            'page': application.sync_namespace.last_page,
            'duration': duration
        })


def page_in(name):
    _page_out()

    application.server['event'].publish('system', {
        'system': {'action': 'page.in', 'page': name}
    })

    # TODO remove below
    application.reporter.report({
        'action': 'page.in',
        'page': name
    })

    application.sync_namespace.last_page = name
    application.sync_namespace.last_page_in_time = datetime.datetime.now()


def start():
    application.sync_namespace.lyrebird_start_time = datetime.datetime.now()
    application.server['event'].publish('system', {
        'system': {'action': 'start'}
    })

    # TODO remove below
    application.reporter.report({
        'action': 'start'
    })


def stop():
    _page_out()
    duration = (datetime.datetime.now() - application.sync_namespace.lyrebird_start_time).total_seconds()
    application.server['event'].publish('system', {
        'system': {
            'action': 'stop',
            'duration': duration
        }
    })

    # TODO remove below
    application.reporter.report({
        'action': 'stop',
        'duration': duration
    })



from flask import request
from lyrebird import application
from flask_restful import Resource

class SettingsApi(Resource):

    def get(self, action):
        if action == 'list':
            resp_dict = {}
            for script_name, script in application.settings.items():
                if not script.inited or not script.template.display:
                    continue
                if script.category_md5 not in resp_dict:
                    resp_dict[script.category_md5] = {
                        'category_md5': script.category_md5,
                        'category': script.category,
                        'scripts': []
                    }
                script_dict = {
                    'name': script.name,
                    'title': script.title,
                    'notice': script.notice,
                    'category': script.category
                }
                resp_dict[script.category_md5]['scripts'].append(script_dict)
            return application.make_ok_response(data=list(resp_dict.values()))
        elif action == 'detail':
            script_name = request.args.get('name')
            if not script_name:
                return application.make_fail_response('Get setting failed, the query \"name\" not found')
            script = application.settings.get(script_name)
            if not script:
                return application.make_fail_response(f'Get setting failed, {script_name} does not exist')
            resp_dict = {
                'name': script.name,
                'title': script.title,
                'notice': script.notice,
                'category': script.category,
                'language': script.language,
                'submitText': script.submit_text,
                'configs': script.getter()
            }
            return application.make_ok_response(data=resp_dict)

    def post(self):
        script_name = request.args.get('name')
        if not script_name:
            return application.make_fail_response('Get setting failed, the query \"name\" not found')

        script = application.settings.get(script_name)
        if not script:
            return application.make_fail_response(f'Get setting failed, {script_name} does not exist')

        resp = script.setter(request.json.get('data'))        

        if resp:
            return application.make_fail_response(str(resp))
        return application.make_ok_response()

    def delete(self):
        script_name = request.args.get('name')
        if not script_name:
            return application.make_fail_response('Get setting failed, the query \"name\" not found')

        script = application.settings.get(script_name)
        if not script:
            return application.make_fail_response(f'Get setting failed, {script_name} does not exist')

        resp = script.restore()        

        if resp:
            return application.make_fail_response(str(resp))
        return application.make_ok_response()



from lyrebird import utils
from lyrebird.mock.dm.jsonpath import jsonpath


class MatchRules:
    def validate_match(rules):
        '''
        { 'request.url': 'pathA/pathB' }

        Empty is allowed
        '''
        if not isinstance(rules, dict):
            return False
        for value in rules.values():
            if not (isinstance(value, (str, int, float, bool)) or value is None):
                return False
        return True

    BOOL_KEY = set(['must', 'must_not'])

    def validate_exists(rules):
        '''
        [ "request.url", "request.data.name" ]

        Empty is allowed
        '''
        if not isinstance(rules, list):
            return False
        for value in rules:
            if not isinstance(value, str):
                return False
        return True

    def validate_query(rules):
        '''
        {
            "match": { ... },
            "exists": [ ... ]
        }

        Empty is allowed
        '''
        if not isinstance(rules, dict):
            return False
        if set(rules.keys()) - set(MatchRules.QUERY_FUNC_MAP.keys()):
            return False

        for query_key, query_rule in rules.items():
            if query_key == 'match':
                if not MatchRules.validate_match(query_rule):
                    return False
            elif query_key == 'exists':
                if not MatchRules.validate_exists(query_rule):
                    return False
        return True

    def validate_bool(rules):
        '''
        {
            "must": { ... },
            "must_not": { ... }
        }
        '''
        if not isinstance(rules, dict):
            return False
        if set(rules.keys()) - set(MatchRules.BOOL_FUNC_MAP.keys()):
            return False

        for search in rules.values():
            if not MatchRules.validate_query(search):
                return False

        return True

    @staticmethod
    def match(flow, rules):
        '''
        1. simple rule
        simple rule is `must.match` in complex rule
        {
            'request.url': 'path/path'
        }

        2. complex rule
        {
            "must": {
                "match": {
                    'request.url': 'path/path'
                }
            },
            "must_not": {
                "match": {
                    "request.url": 'isUpdate'
                },
                "exists": [
                    "request.url"
                ]
            }
        }
        '''

        if not rules:
            return False

        if not isinstance(rules, dict):
            return False

        if MatchRules.is_rule_v1(rules):
            return MatchRules._is_match_rule_v1(flow, rules)

        if MatchRules.is_rule_v2(rules):
            return MatchRules._is_match_rule_v2(flow, rules)

        return False

    @staticmethod
    def is_rule_v1(rules):
        return MatchRules.validate_match(rules)

    @staticmethod
    def is_rule_v2(rules):
        return MatchRules.validate_bool(rules)

    @staticmethod
    def _is_match_rule_v1(flow, rules):
        return MatchRules._query_match(flow, rules)

    @staticmethod
    def _is_match_rule_v2(flow, rules:dict):
        for bool_key, bool_value in rules.items():
            if not bool_value:
                return False
            if bool_key not in MatchRules.BOOL_FUNC_MAP:
                return False
            if not MatchRules.BOOL_FUNC_MAP[bool_key](flow, bool_value):
                return False

        return True

    @staticmethod
    def _bool_must(flow, query:dict, expected=True):
        '''
        query = {
            "match": {
                'request.url': 'path/path'
            }
        }
        '''
        for query_key, query_value in query.items():
            if not query_value:
                return False
            if query_key not in MatchRules.QUERY_FUNC_MAP:
                return False
            if not MatchRules.QUERY_FUNC_MAP[query_key](flow, query_value, expected=expected):
                return False

        return True

    @staticmethod
    def _bool_must_not(flow, query:dict):
        '''
        query = {
            "match": {
                'request.url': 'path/path'
            }
        }
        '''
        return MatchRules._bool_must(flow, query, expected=False)

    @staticmethod
    def _query_exists(flow, exists:dict, expected=True):
        '''
        exists = [
            'request.data'
        ]
        '''
        for exists_key in exists:
            search_res = jsonpath.search(flow, exists_key)
            if bool(search_res) != expected:
                return False
        return True

    @staticmethod
    def _query_match(flow, match:dict, expected=True):
        '''
        match = {
            'request.url': 'path/path'
        }
        '''
        for match_key, match_pattern in match.items():
            targets = MatchRules._get_targets(flow, match_key)
            if targets == []:
                return False
            if not MatchRules._is_targets_pattern_matched(match_pattern, targets, expected=expected):
                return False
        return True

    @staticmethod
    def _get_targets(flow, key):
        search_res = jsonpath.search(flow, key)
        if not search_res:
            return []
        return [s.node for s in search_res]

    @staticmethod
    def _is_targets_pattern_matched(pattern, targets, expected=True):
        for target in targets:
            if utils.TargetMatch.is_match(target, pattern) != expected:
                return False
        return True


    BOOL_FUNC_MAP = {
        'must': _bool_must.__func__,
        'must_not': _bool_must_not.__func__,
    }

    QUERY_FUNC_MAP = {
        'match': _query_match.__func__,
        'exists': _query_exists.__func__
    }



from flask_restful import Resource
from lyrebird.mock import context
from flask import request
from ... import qrcode

LINK_MAX_LENGTH = 2048


class Qrcode(Resource):

    def post(self):
        link = request.json.get('link', '')
        link_length = len(link)
        if link_length > LINK_MAX_LENGTH:
            return context.make_fail_response(f'Current length {link_length} overflows, expected length is less than {LINK_MAX_LENGTH}')
        qrcode_img = qrcode.get_qrcode_img(link)
        if qrcode_img:
            return context.make_ok_response(img=qrcode_img)
        else:
            return context.make_fail_response(f'make qrcode fail! Link is {link}')



import inspect
from lyrebird import application
from .. import checker


class CheckerEventHandler:

    def __call__(self, channel, process=True, *args, **kw):
        def func(origin_func):
            if not checker.scripts_tmp_storage.get(checker.TYPE_EVENT):
                checker.scripts_tmp_storage[checker.TYPE_EVENT] = []
            checker.scripts_tmp_storage[checker.TYPE_EVENT].append({
                'name': origin_func.__name__,
                'origin': origin_func.__code__.co_filename,
                'func': origin_func,
                'channel': channel,
                'process': process
            })
            return origin_func
        return func

    def issue(self, title, message):
        from lyrebird import application
        notice = {
            "title": title,
            "message": message
        }
        self.__class__.check_notice(notice)
        application.server['event'].publish('notice', notice)

    def publish(self, channel, message, *args, **kwargs):
        from lyrebird import application
        if channel == 'notice':
            self.__class__.check_notice(message)
        application.server['event'].publish(channel, message, *args, **kwargs)
    
    @staticmethod
    def check_notice(notice):
        from lyrebird import application
        stack = inspect.stack()
        script_path = stack[2].filename
        script_name = script_path[script_path.rfind('/') + 1:]
        if script_name in application.config.get('event.notice.autoissue.checker', []):
            notice['title'] = f"【Extension】{notice.get('title')}"
            notice['alert'] = False

    @staticmethod
    def register(func_info):
        application.server['event'].subscribe(func_info)

    @staticmethod
    def unregister(func_info):
        application.server['event'].unsubscribe(func_info)

event = CheckerEventHandler()



from pathlib import Path
from lyrebird import utils
from lyrebird.mock.dm.jsonpath import jsonpath
from lyrebird.mock import context
from lyrebird.log import get_logger
from .mock_data_helper import MockDataHelper
from lyrebird.application import config
from lyrebird.config import CONFIG_MOCK_REQUEST_SSR_MOCK_IN_BODY, CONFIG_MOCK_REQUEST_SSR_MOCK_IN_BODY_PARAM, CONFIG_MOCK_REQUEST_SSR_MOCK_BODY_KEY, CONFIG_MOCK_REQUEST_SSR_MOCK_DATA_JSONPATH


logger = get_logger()


class MockHandler:
    """
    根据当前设置数据组的匹配条件,查找对应的mock数据。

    """

    @staticmethod
    def is_handle_ssr(handler_context):
        if not config.get(CONFIG_MOCK_REQUEST_SSR_MOCK_IN_BODY):
            return False

        key = config.get(CONFIG_MOCK_REQUEST_SSR_MOCK_IN_BODY_PARAM, 'is_ssr')
        is_open_ssr = handler_context.flow['request']['query'].get(key)
        if not is_open_ssr:
            return False

        if 'data' in handler_context.flow['request'] and not isinstance(handler_context.flow['request']['data'], dict):
            logger.warning('ssr but request body type is not Object')
            return False

        return True

    def handle(self, handler_context):
        hit_datas_info = context.application.data_manager.get_matched_data_multiple_source(handler_context.flow)
        if not hit_datas_info:
            return

        hit_datas = hit_datas_info['data']
        if len(hit_datas) <= 0:
            return

        # TODO rules of hitting multiple mock data
        hit_data = hit_datas[0]
        activated_group = hit_datas_info['parent']

        is_ssr = MockHandler.is_handle_ssr(handler_context)
        if is_ssr:
            logger.info(
                f'<Mock> <SSR> Hit Group:{activated_group.get("name")} - Data:{hit_data["name"]} \nURL: {handler_context.flow["request"]["url"]}\nGroupID:{activated_group["id"]} DataID:{hit_data["id"]}')

            request_data_update = {}

            resp_key = config.get(CONFIG_MOCK_REQUEST_SSR_MOCK_BODY_KEY, 'lyrebird_mock_response')
            request_data_update[resp_key] = hit_data['response'].get('data', '')

            resp_jsonpath = config.get(CONFIG_MOCK_REQUEST_SSR_MOCK_DATA_JSONPATH)
            if resp_jsonpath:
                request_data_update[resp_key] = utils.flow_str_2_data(request_data_update[resp_key])
                target_response_data = jsonpath.search(request_data_update[resp_key], resp_jsonpath, find_one=True)
                request_data_update[resp_key] = utils.flow_data_2_str(target_response_data)

            if 'data' not in handler_context.flow['request']:
                handler_context.flow['request']['data'] = {}
            handler_context.flow['request']['data'].update(request_data_update)

            handler_context.set_request_edited()
            handler_context.flow['request']['headers']['lyrebird'] = 'mock'

            handler_context.add_flow_action({
                'id': 'mock_ssr',
                'name': Path(__file__).name,
                'group_id': activated_group['id'],
                'mock_id': hit_data['id']
            })
            handler_context.flow['apiDiffConfig'] = hit_data.get('apiDiffConfig')
            return

        logger.info(
            f'<Mock> Hit Group:{activated_group.get("name")} - Data:{hit_data["name"]} \nURL: {handler_context.flow["request"]["url"]}\nGroupID:{activated_group["id"]} DataID:{hit_data["id"]}')
        handler_context.flow['response']['code'] = hit_data['response']['code']
        handler_context.flow['response']['headers'] = {k:v for k,v in hit_data['response']['headers'].items()}
        handler_context.flow['response']['data'] = hit_data['response'].get('data', '')

        handler_context.set_response_edited()
        handler_context.set_response_source_mock()
        handler_context.flow['response']['headers']['isMocked'] = 'True'
        handler_context.flow['response']['headers']['lyrebird'] = 'mock'
        handler_context.add_flow_action({
            'id': 'mock',
            'name': Path(__file__).name,
            'group_id': activated_group['id'],
            'mock_id': hit_data['id']
        })
        handler_context.flow['apiDiffConfig'] = hit_data.get('apiDiffConfig')
        MockDataHelper.read_mock_data(handler_context)



"""
Event bus

Worked as a backgrund thread
Run events handler and background task worker
"""
import os
import sys
import copy
import uuid
import time
import copy
import types
import signal
import pickle
import inspect
import importlib
import functools
import traceback
import setuptools
from concurrent.futures import ThreadPoolExecutor
from lyrebird.base_server import ThreadServer, ProcessServer
from lyrebird.compatibility import prepare_application_for_monkey_patch, monkey_patch_application, monkey_patch_issue, monkey_patch_publish
from lyrebird import application
from lyrebird.mock import context
from lyrebird import log
from pathlib import Path


logger = log.get_logger()
# only report the checker which duration more the 5s
LYREBIRD_METRICS_REPORT_DURSTION = 5000


class InvalidMessage(Exception):
    pass


class Event:
    """
    Event bus inner class
    """

    def __init__(self, event_id, channel, message):
        self.id = event_id
        self.channel = channel
        self.message = message

    def __getstate__(self):
        return pickle.dumps({
            'event_id':self.id,
            'channel':self.channel,
            'message':self.message
            })

    def __setstate__(self, state):
        data = pickle.loads(state)
        self.id = data['event_id']
        self.channel = data['channel']
        self.message = data['message']


def import_func_module(path):
    path = os.path.dirname(path)
    packages = setuptools.find_packages(path)
    for pkg in packages:
        manifest_file = Path(path)/pkg/'manifest.py'
        if not manifest_file.exists():
            continue
        if pkg in sys.modules:
            continue
        sys.path.append(str(Path(path)/pkg))
        spec = importlib.util.spec_from_file_location(pkg, str(Path(path)/pkg/'__init__.py'))
        module = importlib.util.module_from_spec(spec)
        sys.modules[pkg] = module
        spec.loader.exec_module(module)


def import_func_from_file(filepath, func_name):
    name = os.path.basename(filepath)[:-3]
    if name in sys.modules:
        module = sys.modules[name]
    else:
        # 从文件加载模块
        spec = importlib.util.spec_from_file_location(name, filepath)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        sys.modules[name] = module
    return getattr(module, func_name)


def get_func_from_obj(obj, method_name):
    return getattr(obj, method_name)


def get_callback_func(func_ori, func_name):
    if isinstance(func_ori, str):
        return import_func_from_file(func_ori, func_name)
    elif isinstance(func_ori, object):
        return get_func_from_obj(func_ori, func_name)
    else:
        logger.error(f'The source type of method {func_name} is invalid, exception method source: {func_ori}')


def callback_func_run_statistic(callback_fn, args, kwargs, report_info):
    from lyrebird import application
    event_start_time = time.time()
    callback_fn(*args, **kwargs)
    event_end_time = time.time()
    event_duration = (event_end_time - event_start_time) * 1000
    # Report the operation of Event
    # Prevent loop reporting, and only long time event(more than 5s) are reported
    if event_duration < LYREBIRD_METRICS_REPORT_DURSTION:
        return
    if report_info['channel'] == 'lyrebird_metrics':
        return
    if not application.config.get('event.lyrebird_metrics_report', True):
        return
    application.server['event'].publish('lyrebird_metrics', {'lyrebird_metrics': {
        'sender': 'EventServer',
        'action': 'broadcast_handler',
        'duration': event_duration,
        'trace_info': str(report_info['trace_info'])
    }})


class CustomExecuteServer(ProcessServer):
    def __init__(self):
        super().__init__()
        self.event_thread_executor = None
    
    def run(self, async_obj, config, *args, **kwargs):
        self.event_thread_executor = ThreadPoolExecutor(max_workers=async_obj['max_thread_workers'])

        signal.signal(signal.SIGINT, signal.SIG_IGN)

        for _, path in async_obj['plugins']:
            import_func_module(path)

        log_queue = async_obj['logger_queue']
        process_queue = async_obj['process_queue']
        publish_queue = async_obj['publish_queue']

        # monkey_patch is performed on the context content of the process to ensure 
        # that functions of Lyrebird can still be used in the process.

        async_funcs = {}
        async_funcs['publish'] = functools.partial(monkey_patch_publish, publish_queue=publish_queue)
        async_funcs['issue'] = functools.partial(monkey_patch_issue, publish_queue=publish_queue)
        monkey_patch_application(async_obj, async_funcs)

        log.init(config, log_queue)
        self.running = True
        
        while self.running:
            try:
                msg = process_queue.get()
                if not msg:
                    break
                func_ori, func_name, callback_args, callback_kwargs, info = msg
                callback_fn = get_callback_func(func_ori, func_name)
                self.event_thread_executor.submit(callback_func_run_statistic, callback_fn, callback_args, callback_kwargs, info)
            except Exception:
                traceback.print_exc()

    def start(self):
        plugins = application.server['plugin'].plugins
        plugins = [(p_name, plugin.location) for p_name, plugin in plugins.items()]
        self.async_obj['plugins'] = plugins
        self.async_obj['max_thread_workers'] = application.config.get('event.multiprocess.thread_max_worker', 1)
        super().start()


class PublishServer(ThreadServer):
    def __init__(self):
        super().__init__()
        self.publish_msg_queue = application.sync_manager.get_multiprocessing_queue()

    def run(self):
        while self.running:
            try:
                msg = self.publish_msg_queue.get()
                if not msg:
                    break
                event_id, channel, message, args, kwargs = msg
                application.server['event'].publish(channel, message, event_id=event_id, *args, **kwargs)
            except Exception:
                traceback.print_exc()


class EventServer(ThreadServer):

    async_starting = False
    publish_trace_deep = 3

    def __init__(self, no_start = False):
        super().__init__()  
        self.state = {}
        self.pubsub_channels = {}
        # channel name is 'any'. Linstening on all channel
        self.any_channel = []
        self.process_executor_queue = None
        self.event_queue = None
        self.broadcast_executor = None
        self.process_executor = None
        self.publish_server = None
        self.only_report_channel = None
        if not no_start:
            self.only_report_channel = application.config.get('event.only_report_channel', [])
            self.process_executor_queue = application.sync_manager.get_multiprocessing_queue()
            self.event_queue = application.sync_manager.get_queue()
            self.broadcast_executor = ThreadPoolExecutor(thread_name_prefix='event-broadcast-')
            self.process_executor = CustomExecuteServer()
            self.publish_server = PublishServer()

    def broadcast_handler(self, func_info, event, args, kwargs, process_queue=None):
        """

        """
        callback_fn = func_info.get('func')
        is_process = func_info.get('process')

        # Check
        func_sig = inspect.signature(callback_fn)
        func_parameters = list(func_sig.parameters.values())
        if len(func_parameters) < 1 or func_parameters[0].default != inspect._empty:
            logger.error(f'Event callback function [{callback_fn.__name__}] need a argument for receiving event object')
            return
        
        # get enable multiprocess channel list
        multiprocess_channel_list = application.config.get('event.multiprocess.channels', [])

        # Append event content to args
        callback_args = []
        if 'raw' in event.message:
            callback_args.append(event.message['raw'])
        else:
            callback_args.append(event.message)
        # Add channel to kwargs
        callback_kwargs = {}
        if 'channel' in func_sig.parameters:
            callback_kwargs['channel'] = event.channel
        if 'event_id' in func_sig.parameters:
            callback_kwargs['event_id'] = event.id
        # add report info
        info = dict()
        info['trace_info'] = {
            'channel': event.channel,
            'event_id': event.id,
            'callback_fn': callback_fn.__name__,
            'callback_kwargs': str(callback_kwargs)
        }
        info['channel'] = event.channel
        # Execute callback function
        try:
            if EventServer.async_starting and is_process and isinstance(callback_fn, types.FunctionType) and event.channel in multiprocess_channel_list:
                process_queue.put((
                    func_info.get('origin'),
                    func_info.get('name'),
                    callback_args,
                    callback_kwargs,
                    info
                ))
            else:
                callback_func_run_statistic(callback_fn, callback_args, callback_kwargs, info)
        except Exception:
            logger.error(f'Event callback function [{callback_fn.__name__}] error. {traceback.format_exc()}')

    def run(self):
        while self.running:
            try:
                e = self.event_queue.get()
                if not e:
                    break
                # Deep copy event for async event system
                e = copy.deepcopy(e)
                callback_fn_list = self.pubsub_channels.get(e.channel)
                if callback_fn_list:
                    for callback_fn, args, kwargs in callback_fn_list:
                        self.broadcast_executor.submit(self.broadcast_handler, callback_fn, e, args, kwargs, self.process_executor_queue)
                for callback_fn, args, kwargs in self.any_channel:
                    self.broadcast_executor.submit(self.broadcast_handler, callback_fn, e, args, kwargs)
            except Exception:
                # empty event
                traceback.print_exc()

    def async_start(self):
        if not self.publish_server.running:
            self.publish_server.start()
        self.process_namespace = prepare_application_for_monkey_patch()
        self.process_executor.async_obj['process_queue'] = self.process_executor_queue
        self.process_executor.async_obj['process_namespace'] = self.process_namespace
        self.process_executor.async_obj['publish_queue'] = self.publish_server.publish_msg_queue
        self.process_executor.async_obj['eventserver'] = EventServer
        self.process_executor.start()
        application.server['event_process_executor'] = self.process_executor
        EventServer.async_starting = True
        self.publish('system', {'system': {'action': 'event.multiprocess', 'module': 'event_server', 'status': 'READY'}})

    def stop(self):
        self.publish('system', {'name': 'event.stop'})
        time.sleep(1)
        super().stop()
        self.process_executor.stop()
        self.publish_server.stop()

    @staticmethod
    def _check_message_format(message):
        """
        Check if the message content is valid.
        Such as: 'message' value must be a string.
        Other check rules can be added.
        """
        # Check value type of key 'message'
        message_value = message.get('message', 'No message')
        if not isinstance(message_value, str):
            raise InvalidMessage('Value of key "message" must be a string.')

    @staticmethod
    def get_publish_message(channel, message, event_id=None):
        if not event_id:
            # Make event id
            event_id = str(uuid.uuid4())

            # Make sure event is dict
            if not isinstance(message, dict):
                # Plugins send a array list as message, then set this message to raw property
                _msg = {'raw': message}
                message = _msg

            EventServer._check_message_format(message)

            message['channel'] = channel
            message['id'] = event_id
            message['timestamp'] = round(time.time(), 3)

            # Add event sender
            stack = inspect.stack()
            script_path = stack[EventServer.publish_trace_deep].filename
            script_name = script_path[script_path.rfind('/') + 1:]
            function_name = stack[EventServer.publish_trace_deep].function
            sender_dict = {
                "file": script_name,
                "function": function_name
            }
            message['sender'] = sender_dict   
        return (event_id, channel, message)

    def publish(self, channel, message, state=False, event_id=None, *args, **kwargs):
        """
        publish message

        if type of message is dict, set default event information:
            - channel
            - id
            - timestamp
            - sender: if was cantained in message, do not update

        if state is true, message will be kept as state

        """
        event_id, channel, message = EventServer.get_publish_message(channel, message, event_id)

        if channel in self.pubsub_channels or channel not in self.only_report_channel:
            self.event_queue.put(Event(event_id, channel, message))

        # TODO Remove state and raw data
        if state:
            if 'raw' in message:
                self.state[channel] = message['raw']
            else:
                self.state[channel] = message

        # Send event to socket-io
        context.application.socket_io.emit('event', {'id': event_id, 'channel': channel})

        # Send report
        if application.reporter:
            application.reporter.report({
                'action': 'event',
                'channel': channel,
                'event': message
            })

        logger.debug(f'channel={channel} state={state}\nmessage:\n-----------\n{message}\n-----------\n')

    def subscribe(self, func_info, *args, **kwargs):
        """
        Subscribe channel with a callback function
        That function will be called when a new message was published into it's channel

        callback function kwargs:
            channel=None receive channel name
        """
        channel = func_info['channel']
        if 'process' not in func_info:
            func_info['process'] = True
        if channel == 'any':
            self.any_channel.append([func_info, args, kwargs])
        else:
            callback_fn_list = self.pubsub_channels.setdefault(channel, [])
            callback_fn_list.append([func_info, args, kwargs])

    def unsubscribe(self, target_func_info, *args, **kwargs):
        """
        Unsubscribe callback function from channel
        """
        channel = target_func_info['channel']
        if channel == 'any':
            for any_info, *_ in self.any_channel:
                if target_func_info['func'] == any_info['func']:
                    self.any_channel.remove([target_func_info, *_])
        else:
            callback_fn_list = self.pubsub_channels.get(channel)
            for callback_fn_info, *_ in callback_fn_list:
                if target_func_info['func'] == callback_fn_info['func']:
                    callback_fn_list.remove([target_func_info, *_])


class CustomEventReceiver:
    """
    Event Receiver

    Decorator for plugin developer
    Usage:
        event = CustomEventReceiver()

        @event('flow')
        def on_message(data):
            pass
    """

    def __init__(self):
        self.listeners = []

    def __call__(self, channel, object=False, *args, **kw):
        def func(origin_func):
            self.listeners.append(dict(channel=channel, func=origin_func, object=object))
            return origin_func
        return func

    def register(self, event_bus):
        for listener in self.listeners:
            event_bus.subscribe({
                'name': 'CustomEventReceiver',
                'channel': listener['channel'],
                'func': listener['func']
            })

    def unregister(self, event_bus):
        for listener in self.listeners:
            event_bus.unsubscribe({
                'name': 'CustomEventReceiver',
                'channel': listener['channel'],
                'func': listener['func']
            })

    def publish(self, channel, message, *args, **kwargs):
        application.server['event'].publish(channel, message, *args, **kwargs)

    def issue(self, title, message):
        notice = {
            "title": title,
            "message": f'[CustomEventReceiver]: message'
        }
        application.server['event'].publish('notice', notice)



import json
import math
import datetime
import traceback
import time
import copy
from pathlib import Path
from lyrebird import application
from lyrebird import log
from lyrebird.utils import convert_size, convert_size_to_byte, JSONFormat
from lyrebird.base_server import ThreadServer
from lyrebird.mock import context
from sqlalchemy import event, and_
from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, String, Integer, Text, DateTime, create_engine, Table, MetaData
from sqlalchemy.orm.attributes import InstrumentedAttribute
from sqlalchemy.exc import OperationalError


"""
Database server

Worked as a background thread
storage all event message into database
"""

logger = log.get_logger()

Base = declarative_base()


class LyrebirdDatabaseServer(ThreadServer):
    def __init__(self, path=None):
        self.database_uri = None
        self.error_log = []
        self.error_log_threshold = application.config.get('event.db_connection_recover_threshold', 0)
        super().__init__()

        if not path or path.isspace():
            ROOT_DIR = application.root_dir()
            DEFAULT_NAME = 'lyrebird.db'
            self.database_uri = ROOT_DIR/DEFAULT_NAME
        else:
            self.database_uri = Path(path).expanduser().absolute()
        
        # Check whether the current database is broken
        personal_config = application._cm.personal_config
        if str(self.database_uri) in personal_config["event.broken_database_path_list"]:
            if self.database_uri.exists():
                self.database_uri.unlink()
                logger.warning(f"The broken DB has been deleted: {self.database_uri}")
            personal_config["event.broken_database_path_list"].remove(str(self.database_uri))
            application._cm.update_personal_config(personal_config)

        init_engine_success = self.init_engine()
        if not init_engine_success:
            logger.error("Lyrebird database has been broken! Current startup has been stopped, please restart Lyrebird.")
            logger.warning("Restarting will delete the broken database by default, historical events in inspector-pro will be lost, please be careful.")

        # init queue
        self.storage_queue = application.sync_manager.get_queue()

        # subscribe all channel
        application.server['event'].subscribe({
            'name': 'event_receiver',
            'origin': self,
            'channel': 'any',
            'func': self.event_receiver
        })
    
    def auto_alter_tables(self, engine):
        metadata = MetaData()
        tables = {
            table_name: {
                column.name: column for column in Table(table_name, metadata, autoload=True, autoload_with=engine).c
            }
            for table_name in engine.table_names()
        }
        for model_class in Base.__subclasses__():
            table_name = model_class.__table__.name
            if table_name not in tables:
                continue
            table = tables[table_name]
            for attr_name in dir(model_class):
                try:
                    attr = getattr(model_class, attr_name)
                except:
                    logger.warning(f'[Local DB]Cannot get attr:{attr_name} from {model_class}')
                    continue
                if not isinstance(attr, InstrumentedAttribute):
                    continue
                if not hasattr(attr, 'type'):
                    continue
                if not hasattr(attr, 'compile'):
                    continue
                attr_name = attr.name 
                if attr_name in table:
                    continue
                column_type = attr.type.compile(dialect=engine.dialect)
                engine.execute(f'ALTER TABLE {table_name} ADD COLUMN {attr_name} {column_type}')

    def init_engine(self):
        sqlite_path = 'sqlite:///'+str(self.database_uri)+'?check_same_thread=False'

        engine = create_engine(str(sqlite_path))

        # Set pragma on connect
        # https://www.sqlite.org/pragma.html
        event.listen(engine, 'connect', self._fk_pragma_on_connect)

        # Create all tables
        try:
            Base.metadata.create_all(engine)
        except Exception:
            # sqlalchemy.exc.DatabaseError (sqlite3.DatabaseError)
            personal_config = application._cm.personal_config
            personal_config["event.broken_database_path_list"].append(str(self.database_uri))
            application._cm.update_personal_config(personal_config)
            logger.info(f'{traceback.format_exc()}')
            return False
        
        # Create session factory
        session_factory = sessionmaker(bind=engine)
        Session = scoped_session(session_factory)
        self._scoped_session = Session
        self.auto_alter_tables(engine=engine)

        logger.info(f'Init DB engine: {self.database_uri}')
        return True

    def _fk_pragma_on_connect(self, dbapi_con, con_record):
        # https://www.sqlite.org/pragma.html#pragma_journal_mode
        dbapi_con.execute('PRAGMA journal_mode=MEMORY')
        # https://www.sqlite.org/pragma.html#pragma_synchronous
        dbapi_con.execute('PRAGMA synchronous=OFF')

    def event_receiver(self, event, channel=None, event_id=None):
        # event is decoded , which should be encoded when save
        # deepcopy to avoid affecting checker running
        if channel == 'flow':
            event = copy.deepcopy(event)
            application.encoders_decoders.encoder_handler(event['flow'])

        content = json.dumps(event, ensure_ascii=False)
        if isinstance(event, dict):
            message = event.get('message')
        else:
            message = None
        flow = Event(event_id=event_id, channel=channel, content=content, message=message)
        self.storage_queue.put(flow)

    def start(self):
        super().start()

    def run(self):
        session = self._scoped_session()
        while self.running:
            try:
                event = self.storage_queue.get()
                if event is None:
                    break
                session.add(event)
                session.commit()
                context.emit('db_action', 'add event log')
            except OperationalError as e:
                logger.error(f'Save event failed. {traceback.format_exc()}')
                self.error_log.append(e)
                if len(self.error_log) > self.error_log_threshold:
                    logger.warning(f'DB would be reset: {self.database_uri}')
                    self.error_log = []
                    self.reset()
                    session = self._scoped_session()
                else:
                    session.rollback()
            except Exception:
                logger.error(f'Save event failed. {traceback.format_exc()}')
                session.rollback()

    def stop(self):
        super().stop()

    @property
    def session(self):
        return self._scoped_session()

    def get_event(self, channel_rules, offset=0, limit=20, search_str=''):
        search_str_list = [item.strip() for item in search_str.strip().split('+')] if search_str else []
        session = self._scoped_session()
        _subquery = session.query(Event.id).order_by(Event.id.desc())
        if len(channel_rules) > 0:
            _subquery = _subquery.filter(Event.channel.in_(channel_rules))
        if len(search_str_list) > 0:
            _subquery = _subquery.filter(Event.message != None)
            and_cond = []
            for search_str in search_str_list:
                and_cond.append(Event.message.like(f'%%{search_str}%%'))
            _subquery = _subquery.filter(and_(*and_cond))
        _subquery = _subquery.offset(offset).limit(limit).subquery()
        result = session.query(Event).filter(Event.id == _subquery.c.id).all()
        self._scoped_session.remove()
        return result

    def get_page_index_by_event_id(self, event_id, channel_rules, limit=20):

        session = self._scoped_session()
        _subquery = session.query(Event).filter(Event.event_id == event_id).subquery()
        query = session.query(Event.id)
        if len(channel_rules) > 0:
            query = query.filter(Event.channel.in_(channel_rules))
        # Fix bug: event_id in page end, return next page
        # Wrong code: query.filter(Event.id>=subquery.c.id).count()
        result = query.filter(Event.id > _subquery.c.id).count()
        return int(result/limit)

    def get_channel_list(self):
        session = self._scoped_session()
        result = session.query(Event.channel) \
            .group_by(Event.channel) \
            .all()
        self._scoped_session.remove()
        return result

    def get_page_count(self, channel_rules, page_size=20, search_str=''):
        session = self._scoped_session()
        query = session.query(Event.id)
        search_str_list = [item.strip() for item in search_str.strip().split('+')] if search_str else []
        if len(channel_rules) > 0:
            query = query.filter(Event.channel.in_(channel_rules))
        if len(search_str_list) > 0:
            query = query.filter(Event.message != None)
            and_cond = []
            for search_str in search_str_list:
                and_cond.append(Event.message.like(f'%%{search_str}%%'))
            query = query.filter(and_(*and_cond))
        result = query.count()
        self._scoped_session.remove()
        return math.ceil(result / page_size)
    
    def get_database_info(self):
        database_path = str(self.database_uri)
        size = self.database_uri.stat().st_size
        readable_size = convert_size(size)
        database_info = {
            'path': database_path,
            'size': readable_size,
            'oversized': False
        }

        threshold_str = application._cm.config.get('event.file_size_threshold')
        if threshold_str is not None:
            threshold_byte = convert_size_to_byte(threshold_str)
            oversized = threshold_byte and size > threshold_byte
            database_info['threshold'] = threshold_str
            database_info['oversized'] = oversized
        
        return database_info

    def reset(self):
        self.stop()
        self.database_uri.unlink()
        self.init_engine()
        if not self.server_thread.is_alive():
            self.start()
        self.running = True


class Event(Base, JSONFormat):
    __tablename__ = 'event'

    id = Column(Integer, primary_key=True, autoincrement=True)
    channel = Column(String(16), index=True)
    event_id = Column(String(32), index=True)
    content = Column(Text)
    message = Column(Text, default=None)
    _timestamp = Column('timestamp', DateTime(timezone=True), default=datetime.datetime.utcnow)

    @hybrid_property
    def timestamp(self):
        seconds_offset = time.localtime().tm_gmtoff
        return self._timestamp.timestamp() + seconds_offset



import logging
from lyrebird import application
from .base_server import ProcessServer
from logging.handlers import TimedRotatingFileHandler
from colorama import Fore, Style, Back
from collections import namedtuple
from pathlib import Path
import signal
import os
DEFAULT_LOG_PATH = '~/.lyrebird/lyrebird.log'
LOGGER_INITED = False

Color = namedtuple('Color', ['fore', 'style', 'back'])

COLORS = dict(
    NOTICE=Color(fore=Fore.GREEN, style=Style.NORMAL, back=Back.RESET),
    CRITICAL=Color(fore=Fore.WHITE, style=Style.BRIGHT, back=Back.RED),
    ERROR=Color(fore=Fore.RED, style=Style.NORMAL, back=Back.RESET),
    WARNING=Color(fore=Fore.YELLOW, style=Style.NORMAL, back=Back.RESET),
    INFO=Color(fore=Fore.WHITE, style=Style.NORMAL, back=Back.RESET),
    DEBUG=Color(fore=Fore.GREEN, style=Style.NORMAL, back=Back.RESET)
)

process = None
queue_handler = None


def colorit(message, levelname):
    color = COLORS.get(levelname)
    if color:
        return f'{color.fore}{color.style}{color.back}{message}{Style.RESET_ALL}'
    else:
        return message


def info_color(message):
    return f'{Fore.WHITE}{Style.DIM}{Back.RESET}{message}{Style.RESET_ALL}'


class ColorFormater(logging.Formatter):

    def format(self, record: logging.LogRecord):
        s = self.formatTime(record)
        fmt_log = info_color(f'{s} {record.pathname}:{record.lineno}\n') + \
            colorit(f'{record.levelname} {record.msg}', record.levelname)
        return fmt_log


def make_stream_handler():
    color_formater = ColorFormater()
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(color_formater)
    return stream_handler


def make_file_handler(_log_path=None):
    # Set default value
    if not check_path(_log_path):
        _log_path = DEFAULT_LOG_PATH

    # Add pid and thread name in log file
    file_formater = logging.Formatter(
        fmt='%(asctime)s [%(levelname)s] [PID]%(process)s::%(threadName)s (%(pathname)s)  %(message)s'
    )
    log_file = Path(_log_path).expanduser().absolute().resolve()

    file_handler = TimedRotatingFileHandler(log_file, backupCount=1, encoding='utf-8', when='midnight')
    file_handler.setFormatter(file_formater)
    return file_handler


def check_path(path):
    if not path:
        return False
    if not os.path.exists(os.path.dirname(path)):
        return False
    if os.path.isdir(path) or path.split('.')[-1] != 'log':
        return False
    return True


def init(config, log_queue = None):
    global LOGGER_INITED
    global queue_handler
    if LOGGER_INITED:
        return
    
    if not config:
        config = {}
    
    if not log_queue:
        log_server = LogServer()
        log_queue = log_server.queue

    logging.addLevelName(60, 'NOTICE')

    verbose = config.get('verbose', 0)
    if verbose == 0:
        logger_level = logging.ERROR
    elif verbose == 1:
        logger_level = logging.INFO
    elif verbose >= 2:
        logger_level = logging.DEBUG

    queue_handler = logging.handlers.QueueHandler(log_queue)

    for _logger_name in ['lyrebird', 'socketio', 'engineio', 'mock', 'werkzeug', 'flask']:
        _logger = logging.getLogger(_logger_name)
        _logger.addHandler(queue_handler)
        _logger.setLevel(logger_level)
    
    LOGGER_INITED = True


class LogServer(ProcessServer):

    _instance = None

    def __init__(self):
        super().__init__()
        self.queue = application.sync_manager.get_multiprocessing_queue()
    
    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super().__new__(cls, *args, **kwargs)
        return cls._instance

    def run(self, async_obj, config, *args, **kwargs):
        log_queue = async_obj['logger_queue']

        signal.signal(signal.SIGINT, signal.SIG_IGN)

        logging.addLevelName(60, 'NOTICE')

        stream_handler = make_stream_handler()

        log_path = config.get('log')
        file_handler = make_file_handler(log_path)

        verbose = config.get('verbose', 0)
        if verbose == 0:
            logger_level = logging.ERROR
        elif verbose == 1:
            logger_level = logging.INFO
        elif verbose >= 2:
            logger_level = logging.DEBUG

        lyrebird_logger = logging.getLogger('lyrebird')
        lyrebird_logger.addHandler(stream_handler)
        lyrebird_logger.addHandler(file_handler)
        lyrebird_logger.setLevel(logger_level)

        for _logger_name in ['socketio', 'engineio', 'mock', 'werkzeug', 'flask']:
            _logger = logging.getLogger(_logger_name)
            _logger.addHandler(file_handler)
            _logger.setLevel(logger_level)
        
        if log_path and not check_path(log_path):
            lyrebird_logger.warning(f'Illegal log path: {log_path}, log file path have changed to the default path: {DEFAULT_LOG_PATH}')
        
        self.running = True

        while self.running:
            try:
                log = log_queue.get()
                if log is None:
                    break
                logger = logging.getLogger(log.name)
                logger.handle(log)
            except KeyboardInterrupt:
                break
    
    def stop(self):
        super().stop()
        self.queue = None
        logging.shutdown()
        for _logger_name in ['lyrebird', 'socketio', 'engineio', 'mock', 'werkzeug', 'flask']:
            logger = logging.getLogger(_logger_name)
            for handler in logger.handlers[:]:
                logger.removeHandler(handler)
            logger.setLevel(logging.CRITICAL)
    
    def terminate(self):
        super().terminate()
        logging.shutdown()


def get_logger():
    return logging.getLogger('lyrebird')




import asyncio
import re
import traceback
from aiohttp import web, client
from yarl import URL

import sys
import json
from typing import Set
from multidict import CIMultiDict

from lyrebird.mock.extra_mock_server.lyrebird_proxy_protocol import LyrebirdProxyContext
from lyrebird.mock.context import SINGLE_FORWARD_IDENTIFY_HEADERS
from lyrebird import log

logger = None
logger_queue = None
message_queue = None
lb_config = {}
semaphore = None


def is_filtered(context: LyrebirdProxyContext):
    '''
    Copy from lyrebird proxy

    allow list like
    '''
    filters = lb_config.get('proxy.filters', [])
    for _filter in filters:
        if re.search(_filter, context.origin_url):
            return True
    return False


def upgrade_request_report(context: LyrebirdProxyContext):
    if not context.request.headers.get('upgrade'):
        return
    publish_metrics_msg('upgrade_request', {
        'url': context.origin_url,
        'headers': str(context.request.headers)
    })


def make_request_headers(context: LyrebirdProxyContext, is_proxy):
    headers = {k: v for k, v in context.request.headers.items() 
               if k.lower() not in SINGLE_FORWARD_IDENTIFY_HEADERS}
    if is_proxy:
        if 'Lyrebird-Client-Address' in context.request.headers:
            del headers['Lyrebird-Client-Address']
    else:
        if 'Lyrebird-Client-Address' not in context.request.headers:
            headers['Lyrebird-Client-Address'] = context.request.remote
    return headers


async def make_response_header(proxy_resp_headers: dict, context: LyrebirdProxyContext, data=None):
    response_headers = CIMultiDict()
    for k, v in proxy_resp_headers.items():
        if k.lower() == 'content-length':
            if data is not None:
                response_headers.add(k, str(len(data)))
        elif k.lower() == 'host':
            response_headers.add('Host', context.netloc)
        elif k.lower() == 'location':
            response_headers.add('Host', context.netloc)
            response_headers.add(k, v)
        else:
            response_headers.add(k, v)
    return response_headers


async def send_request(context: LyrebirdProxyContext, target_url):
    async with client.ClientSession(auto_decompress=False) as session:
        request: web.Request = context.request
        headers = make_request_headers(context, target_url==context.origin_url)
        request_body = None
        if request.body_exists:
            request_body = request.content
        async with session.request(
            request.method,
            URL(target_url, encoded=True),
            headers=headers,
            data=request_body,
            verify_ssl=False,
            allow_redirects=False,
            raise_for_status=False
        ) as _resp:
            proxy_resp_status = _resp.status
            proxy_resp_headers = _resp.headers
            if 'Transfer-Encoding' in proxy_resp_headers and proxy_resp_headers.get('Transfer-Encoding') == 'chunked':
                response_headers = await make_response_header(proxy_resp_headers, context)
                resp = web.StreamResponse(status=proxy_resp_status, headers=response_headers)
                await resp.prepare(request)
                async for data in _resp.content.iter_any():
                    await resp.write(data)
                await resp.write_eof()
                logger.info(f'Stream Response finished: {proxy_resp_status} {context.origin_url}')
            else:
                proxy_resp_data = await _resp.read()
                response_headers = await make_response_header(proxy_resp_headers, context, proxy_resp_data)
                resp = web.Response(status=proxy_resp_status, body=proxy_resp_data, headers=response_headers)
                logger.info(f'Bytes Response finished: {proxy_resp_status} {context.origin_url}')
    return resp


async def proxy(context: LyrebirdProxyContext):
    logger.info(f'proxy {context.origin_url}')
    return await send_request(context, context.origin_url)


async def forward(context: LyrebirdProxyContext):
    logger.info(f'forward {context.forward_url}')
    try:
        return await send_request(context, context.forward_url)
    except Exception:
        logger.debug(f'Forward request error, it may caused by mock server not ready. URL: {context.forward_url}')


async def req_handler(request: web.Request):
    try:
        global lb_config
        proxy_ctx = LyrebirdProxyContext.parse(request, lb_config)
        upgrade_request_report(proxy_ctx)
        if is_filtered(proxy_ctx):
            # forward to lyrebird
            async with semaphore:
                return await forward(proxy_ctx)
        else:
            # proxy
            return await proxy(proxy_ctx)
    except Exception as e:
        logger.error(f'[Exception] Extra mock server:\n{request.url}\n{traceback.format_exc()}')
        return web.Response(status=500, text=f'{e.__class__.__name__}')


def init_app(config):
    global lb_config
    lb_config = config

    global logger
    log.init(config, logger_queue)
    logger = log.get_logger()

    LyrebirdProxyContext.logger = logger

    app = web.Application()
    app.router.add_route('*', r'/{path:(.*)}', req_handler)

    return app


async def _run_app(config):
    global semaphore
    semaphore = asyncio.Semaphore(20) 
    app = init_app(config)

    port = config.get('extra.mock.port')
    port = port if port else 9999

    try:
        app_runner = web.AppRunner(app, auto_decompress=False)
        await app_runner.setup()

        web_site = web.TCPSite(app_runner, '0.0.0.0', port)
        await web_site.start()

        # app sites list, not display in log.
        # names = sorted(str(s.name) for s in app_runner.sites)
        logger.log(60, f'Extra mock server start on {port}')

        # sleep forever by 1 hour intervals,
        # on Windows before Python 3.8 wake up every 1 second to handle
        # Ctrl+C smoothly
        if sys.platform == "win32" and sys.version_info < (3, 8):
            delay = 1
        else:
            delay = 3600

        while True:
            await asyncio.sleep(delay)
    finally:
        await app_runner.cleanup()


def _cancel_tasks(
    to_cancel: Set["asyncio.Task[Any]"], loop: asyncio.AbstractEventLoop
) -> None:
    if not to_cancel:
        return

    for task in to_cancel:
        task.cancel()

    loop.run_until_complete(asyncio.gather(*to_cancel, return_exceptions=True))

    for task in to_cancel:
        if task.cancelled():
            continue
        if task.exception() is not None:
            loop.call_exception_handler(
                {
                    "message": "unhandled exception during asyncio.run() shutdown",
                    "exception": task.exception(),
                    "task": task,
                }
            )

def publish_init_status(status):
    message_queue.put({
        'type': 'event',
        "channel": "system",
        "content": {
            'system': {
                'action': 'init_module',
                'status': status,
                'module': 'extra_mock'
            }
        }
    })

def publish_metrics_msg(action, info):
    message_queue.put({
        'type': 'event',
        "channel": "lyrebird_metrics",
        "content": {
            'lyrebird_metrics': {
                'sender': 'ExtraMockServer',
                'action': action,
                'trace_info': str(info)
                }
            }
    })

def serve(msg_queue, config, log_queue, *args, **kwargs):
    global logger_queue
    global message_queue
    logger_queue = log_queue
    message_queue = msg_queue

    publish_init_status('READY')
    loop = asyncio.new_event_loop()
    main_task = loop.create_task(_run_app(config))

    try:
        asyncio.set_event_loop(loop)
        loop.run_until_complete(main_task)
    except KeyboardInterrupt:
        publish_init_status('ERROR')
    finally:
        _cancel_tasks({main_task}, loop)
        _cancel_tasks(asyncio.all_tasks(loop), loop)
        loop.run_until_complete(loop.shutdown_asyncgens())
        loop.close()



from lyrebird.log import get_logger
from .server import serve, publish_init_status
from lyrebird.base_server import ProcessServer
import signal

logger = get_logger()


class ExtraMockServer(ProcessServer):

    def run(self, async_obj, config, *args, **kwargs):
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        log_queue = async_obj['logger_queue']
        msg_queue = async_obj['msg_queue']
        serve(msg_queue, config, log_queue, *args, **kwargs)



from setuptools import setup, find_packages
import runpy
import os

here = os.path.abspath(os.path.dirname(__file__))

VERSION = runpy.run_path(
    os.path.join(here, '{{project_name}}', 'version.py')
)['VERSION']

with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='{{package_name}}',
    version=VERSION,
    packages=find_packages(),
    url='',
    author='',
    author_email='',
    long_description=long_description,
    long_description_content_type="text/markdown",
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
    ],
    entry_points={
        'lyrebird_plugin': [
            '{{project_name}} = {{project_name}}.manifest'
        ]
    },
    install_requires=[
        'lyrebird'
    ]
)



"""
Script for mitmdump

Redirect request from proxy server to mock server
"""
import os
import json
from mitmproxy import http
from urllib.parse import urlparse


SERVER_IP = os.environ.get('SERVER_IP')
MOCK_PORT = int(os.environ.get('MOCK_PORT'))
PROXY_PORT = int(os.environ.get('PROXY_PORT'))
PROXY_FILTERS = json.loads(os.environ.get('PROXY_FILTERS'))


def is_websocket_request(flow: http.HTTPFlow) -> bool:
    headers = flow.request.headers
    return (
        headers.get('Upgrade', '').lower() == 'websocket' and
        headers.get('Connection', '').lower() == 'upgrade'
    )


def check_lyrebird_request(flow: http.HTTPFlow):
    parsed_url = urlparse(flow.request.url)
    host = parsed_url.hostname
    port = parsed_url.port
    if host not in ('localhost', '127.0.0.1', SERVER_IP):
        return False
    if not port or port not in (MOCK_PORT, PROXY_PORT):
        return False
    return True


def to_mock_server(flow: http.HTTPFlow):
    raw_url = urlparse(flow.request.url)
    raw_host = raw_url.hostname
    if raw_url.port:
        raw_host += f':{raw_url.port}'
    # mock path 为/mock开头加上原始url
    flow.request.path = '/' + flow.request.url
    # mock scheme 统一为http
    flow.request.scheme = 'http'
    # mock server port
    flow.request.port = PROXY_PORT
    # mock server ip
    flow.request.host = '127.0.0.1'
    # device real ip
    address = flow.client_conn.address[0]
    # 获取的address是IPv6（内嵌IPv4地址表示法），需要获取IPv4地址，需要做以下处理
    if address.startswith('::ffff:'):
        address = address.split('::ffff:')[1]

    flow.request.headers['Lyrebird-Client-Address'] = address
    flow.request.headers['Mitmproxy-Proxy'] = address


def request(flow: http.HTTPFlow):
    if 'mitm.it' in flow.request.url:
        # Support mitm.it
        return
    if check_lyrebird_request(flow):
        # Avoid internal requests
        return
    if is_websocket_request(flow):
        # Avoid Websocket connect requests
        return
    to_mock_server(flow)


def responseheaders(flow):
    """
    Enables streaming for all responses.
    This is equivalent to passing `--set stream_large_bodies=1` to mitmproxy.
    """
    if 'mitm.it' in flow.request.url:
        # Support mitm.it
        flow.response.stream = False
        return
    flow.response.stream = True

    command = flow.response.headers.get('Lyrebird-Mitmproxy-Command')
    if command == 'kill':
        flow.kill()



import json
import codecs
import traceback
import lyrebird
from pathlib import Path
from lyrebird import log
from lyrebird import application
from lyrebird.mock import context

logger = log.get_logger()

class NoticeCenter():

    def __init__(self):
        self.HISTORY_NOTICE = None
        self.notice_hashmap = {}
        self.notice_list = []
        self.not_remind_list = []
        application.server['event'].subscribe({
            'name': 'new_notice',
            'channel': 'notice',
            'func': self.new_notice
        })
        self.load_history_notice()

    def storage_notice(self, storage_date):
        with codecs.open(self.HISTORY_NOTICE, 'w', 'utf-8') as fp:
            fp.writelines(json.dumps(storage_date, ensure_ascii=False, indent=4))
            fp.flush()

    def notice_hashmap_to_list(self):
        self.notice_list, self.not_remind_list = [], []
        for notice_key, notice in self.notice_hashmap.items():
            notice_info_lastone = notice.get('noticeList')[-1].copy()
            notice_info_lastone.update(
                {
                    'key': notice_key,
                    'count': len(notice.get('noticeList'))
                }
            )
            if notice.get('alert'):
                self.notice_list.append(notice_info_lastone)
            else:
                self.not_remind_list.append(notice_info_lastone)
        self.notice_list.sort(key=lambda x:x['timestamp'], reverse=True)
        self.not_remind_list.sort(key=lambda x:x['timestamp'], reverse=True)

    def update_frontend(self):
        """

        emit data and update frontend

        """
        self.notice_hashmap_to_list()
        context.application.socket_io.emit('update')

    def load_history_notice(self):
        """

        load history notice

        """
        ROOT = Path(lyrebird.APPLICATION_CONF_DIR)
        DEFAULT_FILENAME = 'notice.json'
        self.HISTORY_NOTICE = ROOT/DEFAULT_FILENAME
        if self.HISTORY_NOTICE.exists() and self.HISTORY_NOTICE.is_file():
            with codecs.open(self.HISTORY_NOTICE, 'r', 'utf-8') as f:
                try:
                    self.notice_hashmap = json.load(f)
                    self.update_frontend()
                except Exception:
                    self.notice_hashmap = {}
                    logger.error('Load history notice fail!')
                    traceback.print_exc()

    def new_notice(self, msg):
        """

        display new notice
        msg: message dict


        """
        unique_key = msg.get('title')
        alert = msg.get('alert', True)
        if self.notice_hashmap.get(unique_key):
            self.notice_hashmap[unique_key]['noticeList'].append(msg)
            if self.notice_hashmap[unique_key].get('alert'):
                context.application.socket_io.emit('alert', msg)
        else:
            self.notice_hashmap.update(
                {
                    unique_key: {
                        'alert': alert,
                        'noticeList': [msg]
                    }
                }
            )
            context.application.socket_io.emit('alert', msg)

        self.update_frontend()
        self.storage_notice(self.notice_hashmap)

    def change_notice_status(self, unique_key, status):
        """

        change notice alert status, disable notification alert when status is false
        unique_key: str, message unique key
        status: bool, message status


        """
        target_notice = self.notice_hashmap.get(unique_key)
        target_notice['alert'] = status

        self.update_frontend()
        self.storage_notice(self.notice_hashmap)

    def delete_notice(self, unique_key):
        """

        drop notice group in notice hashmap
        unique_key: str, message unique key

        """
        self.notice_hashmap.pop(unique_key)
        self.update_frontend()
        self.storage_notice(self.notice_hashmap)



import os
import json
import shutil
import inspect
import lyrebird
import importlib
import traceback

from lyrebird import application
from lyrebird.log import get_logger
from copy import deepcopy
from types import FunctionType
from pathlib import Path
from lyrebird.base_server import StaticServer
from .settings_template import SettingsTemplate

logger = get_logger()

class Settings:

    def __init__(self, script):
        self.script = script
        self.name = f'{self.script.stem}'
        self.inited = False
        self.template = None

    def __getattr__(self, name):
        if self.template and hasattr(self.template, name):
            return getattr(self.template, name)
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}'")

    def load(self):
        try:
            module_loader = importlib.util.spec_from_file_location(self.name, self.script)
            module = importlib.util.module_from_spec(module_loader)
            module_loader.loader.exec_module(module)
        except Exception as e:
            logger.error(f'Setting item load failed, file path:{str(self.script)}\n {traceback.format_exc()} \n {e}')
        
        for name, obj in inspect.getmembers(module):
            if inspect.isclass(obj) and issubclass(obj, SettingsTemplate) and obj != SettingsTemplate:
                try:
                    template = obj()
                    self.template = template
                    break
                except Exception as e:
                    logger.error(f'Setting item init failed, file path:{self.name}\n {traceback.format_exc()} \n {e}')
        if self.template:
            self.inited = True

    def getter(self):
        if not self.inited:
            return f'Setting item operate is blocked, because it load failed, item name:{self.name}'
        res = {}
        try:
            res = self.template.getter()
        except Exception as e:
            logger.error(f'Setting item getter failed, item name:{self.name}\n {traceback.format_exc()} \n {e}')
        finally:
            return res

    def setter(self, data):
        if not self.inited:
            return f'Setting item operate is blocked, because it load failed, item name:{self.name}'
        res = ''
        try:
            res = self.template.setter(data)
        except Exception as e:
            res = f'Setting item setter failed, item name:{self.name}\n {traceback.format_exc()} \n {e}'
            logger.error(res)

        application.server['event'].publish('lyrebird_metrics', {'lyrebird_metrics': {
            'sender': 'SettingsManager',
            'action': 'setter',
            'trace_info': json.dumps({'name':self.name, 'param':data, 'res':res})
        }})
        return res

    def restore(self):
        if not self.inited:
            return f'Setting item operate is blocked, because it load failed, item name:{self.name}'
        res = ''
        try:
            res = self.template.restore()
        except Exception as e:
            res = f'Setting item restore failed, item name:{self.name}\n {traceback.format_exc()} \n {e}'
            logger.error(res)

        application.server['event'].publish('lyrebird_metrics', {'lyrebird_metrics': {
            'sender': 'SettingsManager',
            'action': 'restore',
            'trace_info': json.dumps({'name':self.name, 'res':res})
        }})
        return res

    def load_finished(self):
        if not self.inited:
            return f'Setting item operate is blocked, because it load failed, item name:{self.name}'
        res = ''
        try:
            res = self.template.load_finished()
        except Exception as e:
            res = f'Setting item load_finished failed, item name:{self.name}\n {traceback.format_exc()} \n {e}'
            logger.error(res)
        return res

    def load_prepared(self):
        if not self.inited:
            return f'Setting item operate is blocked, because it load failed, item name:{self.name}'
        res = ''
        try:
            res = self.template.load_prepared()
        except Exception as e:
            res = f'Setting item load_prepared failed, item name:{self.name}\n {traceback.format_exc()} \n {e}'
            logger.error(res)
        return res

    def destory(self):
        if not self.inited:
            return f'Setting item operate is blocked, because it load failed, item name:{self.name}'
        try:
            self.template.destory()
        except Exception as e:
            logger.error(f'Setting item destory failed, item name:{self.name}\n {traceback.format_exc()} \n {e}')


class SettingsManager(StaticServer):

    def __init__(self):
        self.SCRIPTS_DIR_TEMPLATE = Path(lyrebird.APPLICATION_CONF_DIR)/'settings'
        self.PERSONAL_CONFIG_PATH = Path(lyrebird.APPLICATION_CONF_DIR)/'personal_conf.json'
        self.EXAMPLE_DIR = Path(__file__).parent.parent/'examples'/'settings'
        self.configs = deepcopy(application._cm.personal_config)
        self.settings = application.settings
    
    def stop(self):
        for name, setting in self.settings.items():
            setting.destory()
    
    def load_finished(self):
        for name, setting in self.settings.items():
            setting.load_finished()

    def load_prepared(self):
        for name, setting in self.settings.items():
            setting.load_prepared()

    def load_settings(self):
        scripts_list = self.get_settings_list()
        if not scripts_list:
            return

        switch_conf = application.config.get('settings.switch', [])
        for script in scripts_list:
            if script.name not in switch_conf:
                continue
            if script.inited:
                continue

            script.load()
            if script.inited:
                self.settings[script.name] = script
        
        self.load_prepared()

    def get_settings_list(self):
        workspace_str = application.config.get('settings.workspace')

        if workspace_str:
            workspace = Path(workspace_str)
            if not workspace.expanduser().exists():
                logger.error(f'Settings scripts dir {workspace_str} not found!')
                return
            workspace_iterdir = self.get_iterdir_python(workspace)
            if not workspace_iterdir:
                logger.warning(f'No settings script found in dir {workspace_str}')
                return
        else:
            workspace = Path(self.SCRIPTS_DIR_TEMPLATE)
            workspace.mkdir(parents=True, exist_ok=True)
            workspace_iterdir = self.get_iterdir_python(workspace)
            if not workspace_iterdir:
                self.copy_example_scripts()

        return workspace_iterdir

    @staticmethod
    def get_iterdir_python(path):
        path = Path(path)
        end_str = '.py'
        scripts_list = [Settings(i) for i in path.iterdir() if i.suffix == end_str]
        return scripts_list

    def copy_example_scripts(self):
        for example in self.EXAMPLE_DIR.iterdir():
            if not example.name.endswith('.py'):
                continue
            dst = self.SCRIPTS_DIR_TEMPLATE / example.name
            shutil.copy(example, dst)

    def write_config(self, obj:SettingsTemplate, data={}):
        if data and not isinstance(data, dict):
            return
        template_key = f'settings.{obj.name}'
        template = {
            template_key: self.configs.get(template_key, {})
        }
        if data:
            template[template_key].update(data)
        self.configs.update(template)
        application._cm.update_personal_config(template)

    def get_config(self, obj:SettingsTemplate):
        template_key = f'settings.{obj.name}'
        return self.configs.get(template_key, {})



import re
import json
from hashlib import md5
from lyrebird import application, get_logger
from lyrebird.settings import SettingsTemplate

logger = get_logger()

DEFAULT_WHITE_LIST = []
DEFAULT_BLACK_LIST = []


class WhiteListSettings(SettingsTemplate):

    def __init__(self):
        super().__init__()
        self.display = True
        self.name = 'proxy_white_list'
        self.title = 'Proxy Whitelist and Blacklist Settings'
        self.notice = 'Controls the requests entering Lyrebird proxy logic. Requests filtered out cannot use Mock, Checker, Modifier, etc.'
        self.submit_text = 'Submit'
        self.language = 'en'
        self.category = 'Request Proxy'
        self.category_md5 = md5(self.category.encode(encoding='UTF-8')).hexdigest()
        self.switch = False
        self.configs = {}
        self.ori_filters = application.config.get('proxy.filters', [])
        self.is_simple_url = re.compile(r'^[a-zA-Z0-9./:_-]+$')
        self.is_balck_and_white = re.compile(r'(?=^\(\?=.*\))(?=.*\(\?!.*\)$)')
        self.is_balck = re.compile(r'^\(\?=.*\)$')
        self.is_white = re.compile(r'^\(\?!.*\)$')

    def getter(self):
        white = None
        url_black = None
        suffix_black = None
        regular = None
        if self.configs:
            white = self.configs.get('white_list', [])
            url_black = self.configs.get('black_list', [])
            suffix_black = self.configs.get('black_suffix', [])
            regular = self.configs.get('regular_list', [])
        else:
            filters = self.get_config_by_application()
            white = filters['white']
            regular = filters['regular']
            url_black, suffix_black = self.get_suffix_black_list(filters['black'])
        proxy_white_list_switch = {
            'name': 'proxy_white_list_switch',
            'title': 'Configuration Switch',
            'subtitle': 'Switch for this configuration to take effect. When turned off, it reverts to the default configuration.',
            'category': 'bool',
            'data': self.switch
        }
        white_list = {
            'name': 'white_list',
            'title': 'Request Whitelist',
            'subtitle': 'Allows requests with specific text in host and path to use Lyrebird proxy',
            'category': 'list',
            'data': white
        }
        black_list = {
            'name': 'black_list',
            'title': 'Request Blacklist',
            'subtitle': 'Blocks requests with specific text in host and path from using Lyrebird proxy',
            'category': 'list',
            'data': url_black
        }
        black_suffix = {
            'name': 'black_suffix',
            'title': 'File Type Blacklist',
            'subtitle': 'Globally filters requests of specified resource types, such as png, zip, etc.',
            'category': 'list',
            'data': suffix_black
        }
        regular_list = {
            'name': 'regular_list',
            'title': 'Additional Regex',
            'subtitle': 'If the above whitelists and blacklists do not meet the needs, you can write your own regular expressions. Note: The logic between regexes is OR, meaning as long as any regex matches, it will hit the proxy.',
            'category': 'list',
            'data': regular
        }
        return [proxy_white_list_switch, white_list, black_list, black_suffix, regular_list]

    def setter(self, data):
        self.switch = bool(data['proxy_white_list_switch'])
        self.configs = data
        if self.switch:
            self.apply_config(data)
        self.save(data)
        if not self.switch:
            application.config['proxy.filters'] = self.ori_filters
            logger.warning(f'application.config is updated by proxy_white_list_switch, \nkey: \nproxy.filters \nvalue: \n{json.dumps(self.ori_filters, ensure_ascii=False, indent=4)}')

    def restore(self):
        self.save({})
        self.switch = False
        self.configs = {}
        application.config['proxy.filters'] = self.ori_filters
        logger.warning(f'application.config is updated by proxy_white_list_switch, \nkey: \nproxy.filters \nvalue: \n{json.dumps(self.ori_filters, ensure_ascii=False, indent=4)}')

    def load_prepared(self):
        self.configs = self.manager.get_config(self).get('data', {})
        self.ori_filters = application.config.get('proxy.filters', [])
        have_config = False
        for name, item in self.configs.items():
            if name == 'proxy_white_list_switch':
                self.switch = bool(item)
            else:
                have_config |= bool(item)
        if have_config and self.switch:
            self.apply_config(self.configs)

    def apply_config(self, data):
        filter_list = []
        new_reg = ''
        white_reg = ''
        black_reg = []
        if data['regular_list']:
            filter_list.extend(data['regular_list'])
        if data['white_list']:
            white_reg = '|'.join(data['white_list'])
        if data['black_list']:
            black_reg.append('|'.join(data['black_list']))
        if data['black_suffix']:
            black_reg.append('|'.join(data['black_suffix']))

        if white_reg:
            new_reg = f'(?=.*({white_reg}))'
        if black_reg:
            new_reg += f'(^(?!.*({"|".join(black_reg)})))'

        if not filter_list and not new_reg:
            new_reg = f'(?=.*({"|".join(DEFAULT_WHITE_LIST)}))'
            new_reg += f'(^(?!.*({"|".join(DEFAULT_BLACK_LIST)})))'

        if new_reg:
            filter_list.append(new_reg)
        application.config['proxy.filters'] = filter_list
        logger.warning(f'application.config is updated by proxy_white_list_switch, \nkey: \nproxy.filters \nvalue: \n{json.dumps(filter_list, ensure_ascii=False, indent=4)}')

    def save(self, data):
        self.manager.write_config(self, {'data': data})

    def split_regex(self, regex):
        white = []
        black = []

        if '(?=' in regex or '(?!' in regex:
            positive_parts = re.findall(r'\(\?=\.\*\(?(.*?)\)?\)', regex)
            negative_parts = re.findall(r'\(\?!\.\*\(?(.*?)\)?\)', regex)

            if positive_parts:
                for part in positive_parts:
                    white.extend(part.split('|'))

            if negative_parts:
                for part in negative_parts:
                    black.extend(part.split('|'))

        white = [p for p in white if p.strip()]
        black = [n for n in black if n.strip()]

        return white, black

    def get_config_by_application(self):
        ori_filters = application.config.get('proxy.filters', [])
        white_list = []
        black_list = []
        regular_list = []
        for pattern in ori_filters:
            if self.is_simple_url.match(pattern):
                white_list.append(pattern)
            # Split into three judgments, limit the beginning and end, avoid complex regex misfires.
            elif self.is_balck_and_white.match(pattern) or self.is_white.match(pattern) or self.is_balck.match(pattern):
                res_white, res_black = self.split_regex(pattern)
                white_list.extend(res_white)
                black_list.extend(res_black)
            else:
                regular_list.append(pattern)

        white_list = list(set(white_list))
        black_list = list(set(black_list))
        regular_list = list(set(regular_list))

        return {
            'white': white_list,
            'black': black_list,
            'regular': regular_list
        }

    def get_suffix_black_list(self, black_list):
        url_black_list = [item for item in black_list if not (item.startswith('.') and item.count('.') == 1)]
        suffix_black_list = [item for item in black_list if item.startswith('.') and item.count('.') == 1]
        return url_black_list, suffix_black_list



import os
import json
import codecs
from pathlib import Path
from urllib.parse import urlparse
from lyrebird.log import get_logger
import uuid
import traceback
from copy import deepcopy


logger = get_logger()


PROP_FILE_NAME = '.lyrebird_prop'


class FileDataAdapter:

    def __init__(self, context_self):
        self.context = context_self
        
        self.root_path: Path = None
        self.display_data_map = {}
        self.LEVEL_SUPER_ACTIVATED = 3

        self.COPY_NODE_NAME_SUFFIX = ' - copy'
        self.root_id = None

        self.add_group_ignore_keys = set(['id', 'type', 'children'])
        self.update_group_ignore_keys = set(['id', 'parent_id', 'type', 'children'])
        self.supported_data_type = ['data', 'json', 'config']
        self.virtual_node_data_type = set(['config']) # TODO Next is `super``
        self.virtual_base_config_id = None

        self.unsave_keys = set()

    def _set_root(self, root_path):
        """
        Set a new mock data root dir
        -----
        DataManager will reload all data from this new root dir.
        And all activited data will be removed from DataManager.
        """
        _root_path = Path(root_path).expanduser()
        if not _root_path.exists():
            _root_path.mkdir(parents=True, exist_ok=True)
        if not _root_path.is_dir():
            raise RootPathIsNotDir(root_path)
        self.root_path = _root_path
        self._reload()

    def _reload(self, open_nodes=[]):
        self.context.id_map = {}
        self._load_prop()
        self._sort_children_by_name()
        self.add_parent()
        self.add_super_by()
        self.root_id = self.context.root['id']

    def _load_prop(self):
        path = self.root_path / PROP_FILE_NAME
        if not path.exists():
            self._save_prop()
        with codecs.open(path) as f:
            _root_prop = json.load(f)
        self.context.root = _root_prop
        self._init_id_map(self.context.root)

    def _init_id_map(self, node):
        if 'id' in node:
            self.context.id_map[node['id']] = node
        if 'children' in node:
            for child in node['children']:
                self._init_id_map(child)

    def _get_tree(self):
        return self.context.root

    def _get_group(self, id_):
        return self.context.id_map.get(id_)

    def _get_group_children(self, id_):
        parent_node = self.context.id_map.get(id_)
        if not parent_node:
            logger.error('IDNotFound, id=' + id_)
            return []
        children = parent_node.get('children', [])
        for child in children:
            # update id_map
            if child['id'] not in self.context.id_map:
                self.context.id_map.update({
                    child['id']: child
                })

            self.add_parent_generator(child, parent_node)
            self.add_super_by(node=child)
        children.sort(key=lambda sub_node: sub_node['name'])
        return children

    def _get_link(self, _id):
        if not self.context.root:
            raise RootNotSet
        node = self.context.id_map.get(_id)
        if not node:
            raise IDNotFound(_id)

        link = node.get('link')
        if link:
            link_node = self.context.id_map.get(_id)
            if not link_node:
                logger.error('link node not found!')
                raise IDNotFound(link_node)
            data = self._get(node['link'])
            data['link'] = data['id']
            data['id'] = link_node['id']
            return data
        
    def _load_group(self, _id):
        data = self._get_link(_id)
        if data:
            return data
        return self.context.id_map.get(_id)

    def _add_group(self, data, **kwargs):
        ignore_key = self.add_group_ignore_keys | self.unsave_keys
        new_group = {k: data[k] for k in data if k not in ignore_key}

        parent_id = data.get('parent_id')
        if parent_id == None:
            parent_node = self.context.root
        else:
            parent_node = self.context.id_map.get(parent_id)

        if not parent_node:
            raise IDNotFound(parent_id)
        if parent_node.get('type') == 'data':
            raise OnlyGroupCanContainAnyOtherObject

        # Add group
        group_id = str(uuid.uuid4())
        new_group.update({
            'id': group_id,
            'type': 'group',
            'children': []
        })
        # New group added in the head of child list
        if 'children' not in parent_node:
            parent_node['children'] = []
        parent_node['children'].insert(0, new_group)
        # Register ID
        self.context.id_map[group_id] = new_group
        # Save prop
        self._save_prop()
        return group_id

    def _update_group(self, data, **kwargs):
        id_ = data['id']
        node = self.context.id_map.get(id_)
        if not node:
            raise IDNotFound(id_)

        # 1. Add new key into node
        update_data = {k: data[k] for k in data if k not in self.update_group_ignore_keys}
        node.update(update_data)

        # 2. Remove deleted key in node
        delete_keys = [k for k in node if k not in data and k not in self.update_group_ignore_keys]
        for key in delete_keys:
            node.pop(key)

        # 3. Update existed value
        for key, value in data.items():
            if key in node:
                node[key] = value
        self._save_prop()

    def _delete(self, _id):
        target_node = self.context.id_map.get(_id)
        if not target_node:
            raise IDNotFound(_id)
        parent_id = target_node.get('parent_id')
        # Remove refs
        if parent_id:
            parent = self.context.id_map.get(parent_id)
            parent['children'].remove(target_node)
        else:
            self.context.root['children'].remove(target_node)
        self._delete_inner(_id)

    def _delete_inner(self, _id):
        target_node = self.context.id_map.get(_id)
        if not target_node:
            raise IDNotFound(_id)
        if 'children' in target_node and len(target_node['children']) > 0:
            for child in target_node['children'][::-1]:
                self._delete(child['id'])
        # Delete from ID mapping
        self.context.id_map.pop(_id)
        # Delete from mock tree
        self._delete_group(_id)
        # Delete from file system
        if target_node['type'] == 'data':
            self._delete_data(_id)

    def _delete_group(self, data):
        self._save_prop()

    def _delete_group_by_query(self, query):
        self._save_prop()
    
    def _batch_delete(self, query):
        node_delete_data_ids = query.get('data', [])
        node_delete_group_ids = query.get('groups', [])
        all_id_list = node_delete_data_ids + node_delete_group_ids
        
        for id_ in all_id_list:
            # Delete from parent
            self._delete_remove_from_parent(id_)

            # Delete from ID mapping
            self.context.id_map.pop(id_)
        
        self._delete_group_by_query({'id': node_delete_group_ids})
        self._delete_data_by_query({'id': node_delete_data_ids})

    def _get_type_hashmap(self, ids):
        type_map = {
            'group': [],
            'data': [],
            'json': [],
            'config': []
        }
        for id_ in ids:
            node = self.context.id_map.get(id_)
            if not node:
                continue
            node_type = node['type']
            if node_type not in type_map:
                continue
            type_map[node_type].append(id_)
        return type_map
    
    def _delete_remove_from_parent(self, id_):
        target_node = self.context.id_map.get(id_)
        if not target_node:
            return
        parent_id = target_node.get('parent_id')
        if not parent_id:
            return
        parent = self.context.id_map.get(parent_id)
        if not parent:
            return
        parent['children'].remove(target_node)

    """
    data
    """
    def _load_data(self, data_id, path=None):
        data = self._get_link(data_id)
        if data:
            return data

        if not path:
            path = self.root_path / data_id
        if not Path(path).exists():
            raise DataNotFound(f'Data file {path}')
        with codecs.open(path) as f:
            return json.load(f)

    def _load_data_by_query(self, query):
        datas = []
        for _id in query.get('id', []):
            datas.append(self._load_data(_id))
        return datas

    def _add_data(self, data, path=None, **kwargs):
        parent_id = kwargs.get('parent_id', None)
        data_type = kwargs.get('type', 'data')
        parent_node = self.context.id_map.get(parent_id)
        if not parent_node:
            raise IDNotFound(parent_id)
        
        if parent_node['type'] != 'group':
            raise OnlyGroupCanContainAnyOtherObject
        # Add data
        if not path:
            path = self.root_path / data['id']
        self._save_data(path, data)
        # Add group
        _data_id = data['id']
        if parent_node:
            data_node = {}
            data_node['id'] = _data_id
            data_node['name'] = data.get('name')
            data_node['type'] = data_type
            data_node['parent_id'] = parent_id
            if parent_node.get('children'):
                parent_node['children'].insert(0, data_node)
            else:
                parent_node['children'] = [data_node]

            # New data added in the head of child list
            logger.debug(f'*** Add to node {data_node}')
            # Update ID mapping
            self.context.id_map[_data_id] = data_node
            self._save_prop()

    def _delete_data(self, _id):
        data_file_path = self.root_path / _id
        os.remove(data_file_path)

    def _delete_data_by_query(self, query):
        for _id in query.get('id', []):
            self._delete_data(_id)

    def _update_data(self, _id, data):
        node = self.context.id_map.get(_id)
        if not node:
            raise IDNotFound(_id)
        if 'name' in data and node['name'] != data['name']:
            node['name'] = data['name']
        self._save_data(self.root_path/data['id'], data)
        self._update_group(node)

    def _paste(self, paste_info, **kwargs):
        group_id = paste_info.get('id')
        type = paste_info.get('type')
        parent_id = paste_info.get('parent_id')
        
        _parent_node = self.context.id_map.get(parent_id)
        _node = self.context.id_map.get(group_id)
        
        if not _parent_node:
            raise IDNotFound(parent_id)
        if not _parent_node.get('children'):
            _parent_node['children'] = []
        
        if type == 'cut':
            _origin_parent = self.context.id_map.get(_node['parent_id'])
            _origin_parent['children'].remove(_node)
            _parent_node['children'].insert(0, _node)
            _node['parent_id'] = parent_id
            _group_id = _node['id']
            self._update_group(_node)
        elif type == 'copy':
            new_name = _node['name'] + self.COPY_NODE_NAME_SUFFIX
            _group_id = self._copy_node(_parent_node, _node, paste_info, name=new_name, origin_name=_node['name'], **kwargs)
        elif type == 'import':
            _node = paste_info.get('node', {})
            _group_id = self._copy_node(_parent_node, _node, paste_info, **kwargs)
        return _group_id
    
    def _copy_node(self, parent_node, node, paste_info, **kwargs):
        if self.is_data_virtual_node(node):
            return
        # Generate new node
        new_node = {}
        new_node.update(node)
        new_node['id'] = str(uuid.uuid4())
        new_node['parent_id'] = parent_node['id']
        new_node['name'] = kwargs['name'] if kwargs.get('name') else new_node['name']
        new_node['children'] = []
        # Register ID
        self.context.id_map[new_node['id']] = new_node
        if new_node['type'] == 'group':
            parent_node['children'].insert(0, new_node)
            self._save_prop()
            kwargs.pop('name') if kwargs.get('name') else None
            kwargs.pop('origin_name') if kwargs.get('origin_name') else None
            for child in node['children']:
                self._copy_node(new_node, child, paste_info, **kwargs)
        elif new_node['type'] == 'data':
            self._copy_data(new_node, node, paste_info, **kwargs)
        elif new_node['type'] == 'json':
            self._copy_data(new_node, node, paste_info, **kwargs)
        return new_node['id']
    
    # TODO: 暂时不支持import
    def _copy_data(self, new_data_node, data_node, paste_info, **kwargs):
        _id = data_node['id']

        if paste_info.get('json'):
            event = paste_info.get('json', {}).get(_id)
            if not event:
                raise DataNotFound
            prop = {k:v for k,v in event.items()}
        elif paste_info.get('path'):
            with codecs.open(Path(paste_info['path'])/_id) as f:
                prop = json.load(f)
        else:
            prop = self._load_data(_id)

        prop['id'] = new_data_node['id']
        prop['name'] = kwargs.pop('name') if kwargs.get('name') else prop['name']
        self._add_data(prop, type=data_node['type'], parent_id=new_data_node['parent_id'])

    def is_data_virtual_node(self, node):
        if node['type'] == 'config':
            if 'link' in node:
                return node['link'] == self.virtual_base_config_id
 
        return False

    def _save_prop(self):
        # TODO: Handle frequent writes
        self._sort_children_by_name()
        prop_writer = PropWriter()
        prop_writer.dict_ignore_key.update(self.unsave_keys)

        # handle link node
        prop_writer.dict_ignore_child_key.add('link')

        prop_str = prop_writer.parse(self.context.root)
        prop_file = self.root_path / PROP_FILE_NAME
        with codecs.open(prop_file, 'w') as f:
            f.write(prop_str)

    def _save_data(self, path, data):
        new_data = deepcopy(data)
        new_data.pop('type', None)
        prop_str = json.dumps(new_data, ensure_ascii=False)
        with codecs.open(path, 'w') as f:
            f.write(prop_str)

    def _get_data_name(self, data):
        name = 'New Data'
        request = data.get('request')
        if not request:
            return name

        url = request.get('url')
        if not url:
            return name

        parsed_url = urlparse(url)
        host = parsed_url.hostname
        path = parsed_url.path
        if path:
            name = path
        elif host:
            name = host
        else:
            name = url[0:100]

        return name

    def _get_data_rule(self, request):
        pattern = 'YOUR-REQUEST-PATH'
        url = request.get('url')
        if not url:
            return {'request.url': f'(?=.*{pattern})'}

        parsed_url = urlparse(url)
        host = parsed_url.netloc
        path = parsed_url.path
        query = parsed_url.query
        if path and host:
            pattern = path
            if query:
                pattern += '\?' 
            elif url.endswith(path):
                pattern += '$'

        elif host:
            pattern = host
            if url.endswith(host):
                pattern += '$'

        else:
            pattern = url
        return {'request.url': f'(?=.*{pattern})'}
    
    def _activate(self, search_id):
        if not self.context.root:
            raise RootNotSet
        _node = self.context.id_map.get(search_id)
        if not _node:
            raise IDNotFound(f'ID:{search_id}')
        node_id_list = self._collect_activate_node(_node, level_lefted=self.LEVEL_SUPER_ACTIVATED)
        ordered_data_id = []
        for node_id in node_id_list:
            activated_node = self.context.id_map.get(node_id)
            ordered_data_id += self._collect_data_in_node(activated_node)
        node_list = self._load_data_by_query({'id': ordered_data_id})
        config = {}
        for node in node_list:
            node_in_map = self.context.id_map.get(node.get('id'))
            if node_in_map['type'] == 'config':
                config = node
        return {
            'config': config,
            'flows': node_list,
            'groupInfo': _node
        }

    def _collect_data_in_node(self, node):
        id_list = []
        if self.is_data_virtual_node(node):
            return []
        if node.get('type') in self.supported_data_type:
            return [node['id']]
        elif node.get('type', '') == 'group':
            if 'children' in node:
                for child in node['children']:
                    id_list.extend(self._collect_data_in_node(child))
        return id_list

    def _collect_activate_node(self, node, level_lefted=1):
        id_list = [node['id']]
        if level_lefted <= 0:
            return id_list
        if not node.get('super_id'):
            return id_list
        if node.get('super_id') == node['id']:
            raise SuperIdCannotBeNodeItself(node['id'])

        _super_id = node.get('super_id')
        _super_node = self.context.id_map.get(_super_id)
        if not _super_node:
            raise IDNotFound(f'Super node ID: {_super_id}')

        id_list.extend(self._collect_activate_node(_super_node, level_lefted=level_lefted-1))
        return id_list

    def _get_activate_group(self, search_id):
        return self.context.id_map.get(search_id)

    def _after_activate(self, **kwargs):
        pass

    def add_parent(self, node=None):
        if not node:
            root_id = self.context.root['id']
            root = self.context.id_map.get(root_id)
            self.add_parent_generator(root)
        else:
            parent_node = self.context.id_map.get(node['parent_id'])
            self.add_parent_generator(node, parent_node=parent_node)
        self.unsave_keys.update(['parent', 'abs_parent_path'])
    
    def add_super_by(self, node=None):
        node_list = [node] if node else self.context.id_map.values()

        for node in node_list:
            if not node.get('super_id'):
                continue
            super_parent_node = self.context.id_map.get(node['super_id'])
            if not super_parent_node:
                continue
            if not super_parent_node.get('super_by'):
                super_parent_node['super_by'] = []
            super_parent_node['super_by'].append({
                'id': node['id'],
                'name': node['name']
            })
        self.unsave_keys.add('super_by')

    def add_parent_generator(self, node, parent_node=None):
        parent_node_parent = parent_node['parent'] if parent_node and parent_node.get('parent') else []
        parent_obj = parent_node_parent + [{
            'id': node['id'],
            'name': node['name'],
            'type': node['type'],
            'parent_id': node['parent_id']
        }]

        abs_parent_path = f"{parent_node['abs_parent_path']}{node['name']}/" if parent_node and parent_node.get('abs_parent_path') else '/'

        node.update({
            'parent': parent_obj,
            'abs_parent_path': abs_parent_path
        })

        if not node.get('children'):
            return

        for child in node['children']:
            self.add_parent_generator(child, node)

    def _sort_children_by_name(self):
        for node_id in self.context.id_map:
            node = self.context.id_map[node_id]
            if 'children' not in node:
                # fix mock data group has no children
                if node['type'] == 'group':
                    node['children'] = []
                continue
            node['children'] = sorted(node['children'], key=lambda sub_node: sub_node['name'])

    def _search(self, search_str, sender=None):
        _matched_group = []
        if search_str:
            if self.context.id_map.get(search_str):
                # search_str is group id
                group = self.context.id_map.get(search_str)
                _matched_group.append(self.make_search_group_item(group))
            else:
                for _id, group in self.context.id_map.items():
                    if group.get('type') == 'group' and search_str.lower() in group.get('name', '').lower():
                        _matched_group.append(self.make_search_group_item(group))
        else:
            for _id, group in self.context.id_map.items():
                if group.get('type') == 'group' and group.get('name'):
                    _matched_group.append(self.make_search_group_item(group))
        logger.info(f'搜索到{len(_matched_group)}个数据组')
        return _matched_group

    def make_search_group_item(self, _group):
        return {
                'id': _group['id'],
                'name': _group['name'],
                'type': _group['type'],
                'parent_id': _group['parent_id'],
                'abs_parent_path': self._get_abs_parent_path(_group),
                'abs_parent_obj': self._get_abs_parent_obj(_group)
            }
    
    def _get_abs_parent_path(self, data, path='/'):
        id = data.get('id')
        if not id:
            return path
       
        node = self.context.id_map.get(id)
        if 'parent_id' in node:
            parent_node = self.context.id_map.get(node['parent_id'])
        
        if not parent_node:
            return path
        current_path = '/' + node['name'] + path
        return self._get_abs_parent_path(parent_node, path=current_path)
    
    def _get_node_parent(self, node):
        if 'parent_id' not in node:
            return None
        parent_node = self.context.id_map.get(node['parent_id'])
        if not parent_node:
            return None
        return parent_node
    
    def _get_abs_parent_obj(self, node, parent_obj=None):
        if parent_obj is None:
            parent_obj = []
        if 'id' not in node:
            return parent_obj
        node_info = self.context.id_map.get(node['id'])
        if not node_info:
            return parent_obj
        parent_obj.insert(0, {
            'id': node_info['id'],
            'name': node_info['name'],
            'type': node_info['type'],
            'parent_id': node_info['parent_id']
        })
        parent_node = self._get_node_parent(node)
        if parent_node is None:
            return parent_obj
        return self._get_abs_parent_obj(parent_node, parent_obj=parent_obj)
    
    def _get_node_prop(self, _id):
        node = self.context.id_map.get(_id)
        ordered_data_id = self._collect_data_in_node(node)
        return {
            'prop': node,
            'childDataIds': ordered_data_id
        }

    # -----
    # Get all data under one id
    # -----

    def _read_data(self, _id):
        node = self.context.id_map.get(_id)
        if not node:
            raise IDNotFound(_id)
        data_array = []
        if node.get('type') == 'data':
            _data = self._load_data(node['id'])
            _data['parent_id'] = node['parent_id']
            data_array.append(_data)
        elif node.get('type') == 'group':
            for child in node['children']:
                data_array.extend(self._read_data(child.get('id')))
        return data_array

class PropWriter:

    def __init__(self):
        self.indent = 0
        self.parsers = {
            'dict': self.dict_parser,
            'list': self.list_parser,
            'int': self.int_parser,
            'str': self.str_parser,
            'bool': self.bool_parser,
            'NoneType': self.none_parser
        }
        self.dict_ignore_key = set()
        self.dict_ignore_child_key = set()

    def parse(self, prop):
        prop_type = type(prop)
        parser = self.parsers.get(prop_type.__name__)
        if not parser:
            raise DumpPropError(f'Not support type {prop_type}')
        return parser(prop)

    def dict_parser(self, val):
        dict_str = '{'
        children = None
        for k, v in val.items():
            if k in self.dict_ignore_key:
                continue
            if k == 'children':
                children = v
                continue
            dict_str += f'"{k}":{self.parse(v)},'
        if children is not None:
            dict_str += self.children_parser(children)
        if dict_str.endswith(','):
            dict_str = dict_str[:-1]
        dict_str += '}'
        return dict_str

    def list_parser(self, val):
        list_str = '['
        for item in val:
            item_str = self.parse(item)
            list_str += item_str + ','
        if list_str.endswith(','):
            list_str = list_str[:-1]
        list_str += ']'
        return list_str

    def int_parser(self, val):
        return f'{val}'

    def str_parser(self, val):
        return json.dumps(val, ensure_ascii=False)

    def bool_parser(self, val):
        return json.dumps(val)

    def none_parser(self, val):
        return 'null'

    def children_parser(self, val):
        self.indent += 1
        children_str = '"children":['
        for child in val:
            if self.dict_ignore_child_key & set(child.keys()):
                continue
            child_str = self.parse(child)
            children_str += '\n' + '  '*self.indent + child_str + ','
        if children_str.endswith(','):
            children_str = children_str[:-1]
        children_str += ']'
        self.indent -= 1
        return children_str


class IDNotFound(Exception):
    pass


class DataNotFound(Exception):
    pass


class DumpPropError(Exception):
    pass

class RootPathIsNotDir(Exception):
    pass


class RootNotSet(Exception):
    pass

class UnsupportedType(Exception):
    pass


class OnlyGroupCanContainAnyOtherObject(Exception):
    pass


class SuperIdCannotBeNodeItself(Exception):
    pass


data_adapter = FileDataAdapter



import uuid
from lyrebird.utils import JSONFormat


PLACEMENT_BOTTOM_LEFT = 'bottom_left'
PLACEMENT_BOTTOM_RIGHT = 'bottom_right'
PLACEMENT_TOP_RIGHT = 'top_right'


class PlainText:
    """
    Status plain text component

    On main UI rendering , Lyrebird will call getText function and display its return on footbar or appbar.

    Attributes:
        - prepend_icon use Material Design icon
        - placement: accept PLACEMENT_BOTTOM_LEFT, PLACEMENT_BOTTOM_RIGHT and PLACEMENT_TOP_RIGHT
        - rank: The larger the data, the closer to the target placement
        - icon: Display in QRCode
    """
    rank = 0
    prepend_icon = None
    placement = PLACEMENT_BOTTOM_LEFT

    def __init__(self):
        self.display = True
        self.id = str(uuid.uuid4())
        if not hasattr(self, 'name'):
            self.name = self.id

    @property
    def type(self):
        return self.__class__.__base__.__name__

    def get_text(self):
        """
        return a string
        """
        pass


class ClickableStatusText:
    """
    Plugin status text-tooltip component

    On main UI rendering , Lyrebird will call getText function and display its return on footbar or appbar.
    If user click that text， Lyrebird will call getMenu function for loading menu items.

    Attributes:
        - prepend_icon use Material Design icon
        - placement: accept PLACEMENT_BOTTOM_LEFT, PLACEMENT_BOTTOM_RIGHT and PLACEMENT_TOP_RIGHT
        - rank: The larger the data, the closer to the target placement
        - icon: Display in QRCode
    """

    rank = 0
    prepend_icon = None
    placement = PLACEMENT_BOTTOM_LEFT

    def __init__(self):
        self.id = str(uuid.uuid4())
        if not hasattr(self, 'name'):
            self.name = self.id

    @property
    def type(self):
        return self.__class__.__base__.__name__

    def get_text(self):
        """
        return a string
        """
        pass

    def get_menu(self):
        """
        return a list of MenuItem

        e.g.
        return [TextMenuItem('Hello'), TextMenuItem('World')]
        """
        pass

    def get_detail(self):
        return [menu_item.json() for menu_item in self.get_menu()]

    def json(self):
        info = JSONFormat.json(self)
        info.update({
            'text': self.get_text()
        })
        return info


class MenuItem:

    def __init__(self, src):
        self.id = str(uuid.uuid4())
        self.type = self.__class__.__name__
        self.src = src

    def json(self):
        return {'id': self.id, 'type': self.type, 'src': self.src}


class ImageMenuItem(MenuItem):
    """
    The src argument support URL and base64 data string
    """
    pass


class TextMenuItem(MenuItem):
    """
    src: text
    """
    pass


class LinkMenuItem(MenuItem):
    """
    src: link
    """
    pass


class Selector:
    """
    Plugin status select component

    On main UI rendering , Lyrebird will call getMenu function and display its return on footbar or appbar.

    Attributes:
        - type: 
        - prepend_icon: use Material Design icon
        - placement: accept PLACEMENT_BOTTOM_LEFT, PLACEMENT_BOTTOM_RIGHT and PLACEMENT_TOP_RIGHT
        - rank: The larger the data, the closer to the target placement
    """

    rank = 0
    prepend_icon = None
    placement = PLACEMENT_BOTTOM_LEFT

    def __init__(self):
        self.id = str(uuid.uuid4())
        if not hasattr(self, 'name'):
            self.name = self.id

    @property
    def type(self):
        return self.__class__.__base__.__name__

    def get_menu(self):
        """
        return a dict
        {
            'selectedIndex': selected_index,
            'selected': selected_api,
            'allItem': [{
                'text': 'Display text',
                'api': ''
            }]
        }
        """
        pass

    def get_detail(self):
        return self.get_menu()

    def json(self):
        info = JSONFormat.json(self)
        info.update({
            'src': self.get_menu()
        })
        return info



"""
Base threading server class
"""

import inspect
from threading import Thread
from multiprocessing import Process
from lyrebird import application


def check_process_server_run_function_compatibility(function):
    # Check whether the run method is an old or new version by params.
    if len(inspect.signature(function).parameters) == 4 and list(inspect.signature(function).parameters.keys())[0] == 'async_obj':
        return True
    else:
        return False


service_msg_queue = None


class ProcessServer:
    def __init__(self):
        self.server_process = None
        self.running = False
        self.name = None
        self.event_thread = None
        self.async_obj = {}
        self.args = []
        self.kwargs = {}

    def run(self, async_obj, config, *args, **kwargs):
        '''
        async_obj is a dict 
        used to pass in all objects used for synchronization/communication between multiple processes
        Usually msg_queue, config and log_queue is included
            msg_queue:
                message queue for process server and main process

                #1. Send event to main process,
                {
                    "type": "event",
                    "channel": "",
                    "content": {}
                }

                #2. Send message to frontend
                support channel: msgSuccess msgInfo msgError
                {
                    "type": "ws",
                    "channel": "",
                    "content": ""
                }

            config:
                lyrebird config dict

            log_queue:
                send log msg to logger process
        '''
        pass

    def start(self):
        if self.running:
            return
        
        from lyrebird.log import get_logger
        logger = get_logger()

        global service_msg_queue
        if service_msg_queue is None:
            service_msg_queue = application.sync_manager.get_multiprocessing_queue()
        config = application._cm.config
        logger_queue = application.server['log'].queue

        # run method has too many arguments. Merge the msg_queue, log_queue and so on into async_obj
        # This code is used for compatibility with older versions of the run method in the plugin
        # This code should be removed after all upgrades have been confirmed
        if check_process_server_run_function_compatibility(self.run):
            self.async_obj['logger_queue'] = logger_queue
            self.async_obj['msg_queue'] = service_msg_queue
            self.server_process = Process(group=None, target=self.run,
                                        args=[self.async_obj, config, self.args],
                                        kwargs=self.kwargs,
                                        daemon=True)
        else:
            logger.warning(f'The run method in {type(self).__name__} is an old parameter format that will be removed in the future')
            self.server_process = Process(group=None, target=self.run,
                                        args=[service_msg_queue, config, logger_queue, self.args],
                                        kwargs=self.kwargs,
                                        daemon=True)
        self.server_process.start()
        self.running = True

    def stop(self):
        self.running = False
    
    def terminate(self):
        if self.server_process:
            self.server_process.terminate()
            self.server_process.join()
            self.server_process = None


class ThreadServer:

    def __init__(self):
        self.server_thread = None
        self.running = False
        self.name = None

    def start(self, *args, **kwargs):
        if self.running:
            return
        self.running = True
        self.server_thread = Thread(target=self.run, name=self.name, args=args, kwargs=kwargs)
        self.server_thread.start()

    def stop(self):
        self.running = False
        # TODO terminate self.server_thread
    
    def terminate(self):
        pass

    def run(self):
        """
        Server main function
        """
        pass


class StaticServer:

    def start(self, *args, **kwargs):
        pass

    def stop(self):
        pass

    def terminate(self):
        pass


class MultiProcessServerMessageDispatcher(ThreadServer):

    def run(self):
        global service_msg_queue
        if service_msg_queue is None:
            service_msg_queue = application.sync_manager.get_multiprocessing_queue()
        emit = application.server['mock'].socket_io.emit
        publish = application.server['event'].publish

        while self.running:
            msg = service_msg_queue.get()
            if msg is None:
                break
            type = msg.get('type')
            if type == 'event':
                channel = msg.get('channel')
                event = msg.get('content')
                if channel and event:
                    publish(channel, event)
            elif type == 'ws':
                ws_channel = msg.get('channel')
                ws_msg = msg.get('content', '')
                if ws_channel:
                    emit(ws_channel, ws_msg)
            else:
                pass



from collections import OrderedDict
from . import content_encoding, content_type
from lyrebird.utils import CaseInsensitiveDict
from lyrebird.mock.context import LYREBIRD_UNPROXY_HEADERS
import json

origin2flow_handlers = OrderedDict({
    'Content-Encoding': content_encoding,
    'Content-Type': content_type
})

flow2origin_handlers = OrderedDict({
    'Content-Type': content_type,
    'Content-Encoding': content_encoding
})


class DataHelper:

    @staticmethod
    def origin2flow(origin_obj, output=None, chain=None):
        if not origin_obj:
            return

        _data = origin_obj.data
        if not _data:
            return

        if 'Proxy-Raw-Headers' in origin_obj.headers:
            _origin_headers = json.loads(origin_obj.headers['Proxy-Raw-Headers'])
            raw_headers = CaseInsensitiveDict(_origin_headers)
        # Read raw headers, support the request from extra mock 9999 port
        elif hasattr(origin_obj, 'environ') and '_raw_header' in origin_obj.environ:
            raw_headers = CaseInsensitiveDict(origin_obj.environ['_raw_header'])
            for key in LYREBIRD_UNPROXY_HEADERS:
                if key in raw_headers:
                    del raw_headers[key]
        else:
            raw_headers = origin_obj.headers

        for headers_key, func in origin2flow_handlers.items():
            headers_val = raw_headers.get(headers_key, '')
            _data = func.origin2flow(headers_val, _data, chain=chain)

        if output:
            output['data'] = _data
        else:
            return _data

    @staticmethod
    def flow2origin(flow_obj, output=None, chain=None):
        _data = flow_obj.get('data')
        if _data is None:
            return

        if chain:
            for func in chain[::-1]:
                _data = func.flow2origin(_data)
        else:
            for headers_key, func in flow2origin_handlers.items():
                headers_val = flow_obj['headers'].get(headers_key, '')
                _data = func.flow2origin(headers_val, _data)

        if output:
            output.data = _data
        else:
            return _data

    @staticmethod
    def origin2string(origin_obj, output=None):
        if not origin_obj:
            return

        _data = origin_obj.data
        if not _data:
            return

        if 'Proxy-Raw-Headers' in origin_obj.headers:
            _origin_headers = json.loads(origin_obj.headers['Proxy-Raw-Headers'])
            raw_headers = CaseInsensitiveDict(_origin_headers)
        # Read raw headers, support the request from extra mock 9999 port
        elif hasattr(origin_obj, 'environ') and '_raw_header' in origin_obj.environ:
            raw_headers = CaseInsensitiveDict(origin_obj.environ['_raw_header'])
            for key in LYREBIRD_UNPROXY_HEADERS:
                if key in raw_headers:
                    del raw_headers[key]
        else:
            raw_headers = origin_obj.headers

        for headers_key, func in origin2flow_handlers.items():
            headers_val = raw_headers.get(headers_key, '')
            _data = func.origin2string(headers_val, _data)

        if output:
            output['data'] = _data
        else:
            return _data



import sys
import platform
import importlib.util
from lyrebird import log
from functools import wraps
from inspect import signature
from multiprocessing import managers
from multiprocessing.managers import Namespace

logger = log.get_logger()


application_white_map = {
    'config',
    '_cm'
}


context_white_map = {
    'application.data_manager.activated_data',
    'application.data_manager.activated_group',
    'application.data_manager.async_activate_group'
}


decorator_compat_code = '''
def jit(*args, **kwargs):
    def decorator(func):
        print(f"{func.__name__} run in a cover function")
        return func
    return decorator
'''

PYTHON_MIN_VERSION = (3, 8, 0)
PYTHON_MAX_VERSION = (3, 13, float('inf'))


'''
Python3.8 and earlier do not support managing SyncManger managed objects
github original issue: https://github.com/python/cpython/pull/4819
'''
orig_AutoProxy = managers.AutoProxy

@wraps(managers.AutoProxy)
def AutoProxy(*args, incref=True, manager_owned=False, **kwargs):
    # Create the autoproxy without the manager_owned flag, then
    # update the flag on the generated instance. If the manager_owned flag
    # is set, `incref` is disabled, so set it to False here for the same
    # result.
    autoproxy_incref = False if manager_owned else incref
    proxy = orig_AutoProxy(*args, incref=autoproxy_incref, **kwargs)
    proxy._owned_by_manager = manager_owned
    return proxy


def compat_async_manager_to_python_3_8():
    if "manager_owned" in signature(managers.AutoProxy).parameters:
        return

    logger.debug("Patching multiprocessing.managers.AutoProxy to add manager_owned")
    managers.AutoProxy = AutoProxy

    # re-register any types already registered to SyncManager without a custom
    # proxy type, as otherwise these would all be using the old unpatched AutoProxy
    SyncManager = managers.SyncManager
    registry = managers.SyncManager._registry
    for typeid, (callable, exposed, method_to_typeid, proxytype) in registry.items():
        if proxytype is not orig_AutoProxy:
            continue
        create_method = hasattr(managers.SyncManager, typeid)
        SyncManager.register(
            typeid,
            callable=callable,
            exposed=exposed,
            method_to_typeid=method_to_typeid,
            create_method=create_method,
        )


def import_compat_util(module_name:str, module_content:list):
    module_spec = importlib.util.spec_from_loader(module_name, loader=None, origin='string', is_package=False)
    module_obj = importlib.util.module_from_spec(module_spec)
    for code in module_content:
        exec(code, module_obj.__dict__)
    sys.modules[module_name] = module_obj


def compat_redis_check():
    try:
        import redis
    except Exception as e:
        logger.error(f'redis import failed. Please check that the library is installed correctly in your python environment')
        return False
    return True


def compat_python_version_check():
    version = platform.python_version_tuple()
    major = int(version[0])
    minor = int(version[1])
    minor_minor = int(version[2])
    if major < PYTHON_MIN_VERSION[0] or \
    (major == PYTHON_MIN_VERSION[0] and minor < PYTHON_MIN_VERSION[1]) or \
    (major == PYTHON_MIN_VERSION[0] and minor == PYTHON_MIN_VERSION[1] and minor_minor == PYTHON_MIN_VERSION[2]):
        msg = (
            'The python version is too early. Please use Python version '
            f'{PYTHON_MIN_VERSION[0]}.{PYTHON_MIN_VERSION[1]}.{PYTHON_MIN_VERSION[2]} or later.'
        )
        logger.error(msg)
        return False
    if major > PYTHON_MAX_VERSION[0] or \
    (major == PYTHON_MAX_VERSION[0] and minor > PYTHON_MAX_VERSION[1]) or \
    (major == PYTHON_MAX_VERSION[0] and minor == PYTHON_MAX_VERSION[1] and minor_minor > PYTHON_MAX_VERSION[2]):
        msg = (
            'python version is too high, Lyrebird is not supported,'
            'the current Lyrebird support version is '
            f'{PYTHON_MAX_VERSION[0]}.{PYTHON_MAX_VERSION[1]}.{"x" if isinstance(PYTHON_MAX_VERSION[2], float) else PYTHON_MAX_VERSION[2]}.'
        )
        logger.warning(msg)
    return True


def prepare_application_for_monkey_patch() -> Namespace:
    from lyrebird import application, context
    namespace = application.sync_manager.get_namespace()
    namespace.application = ProcessApplicationInfo(application, application_white_map)
    namespace.context = ProcessApplicationInfo(context, context_white_map)
    namespace.queue = application.server['event'].event_queue
    return namespace


def monkey_patch_application(async_obj, async_funcs=None, async_values=None):
    import lyrebird
    from lyrebird.event import EventServer
    from lyrebird import event

    # TODO Short-term fix, Optimized to general logic
    EventServer.publish_trace_deep -= 1

    msg_queue = async_obj['msg_queue']
    process_namespace = async_obj['process_namespace']
    
    lyrebird.application = process_namespace.application
    lyrebird.application.config = process_namespace.application._cm.config
    lyrebird.context = process_namespace.context

    monkey_patch_datamanager()

    if async_funcs:
        checker_event_server = EventServer(True)
        checker_event_server.event_queue = msg_queue
        lyrebird.application['server'] = ProcessApplicationInfo()
        lyrebird.application.server['event'] = checker_event_server
        checker_event_server.__class__.publish = async_funcs['publish']
        event.__class__.publish = async_funcs['publish']
        event.__class__.issue = async_funcs['issue']


def monkey_patch_datamanager():
    from lyrebird import context
    context.application.data_manager.activate_group = context.application.data_manager.async_activate_group


def monkey_patch_publish(channel, message, publish_queue, *args, **kwargs):
    from lyrebird.event import EventServer
    from lyrebird.checker.event import CheckerEventHandler
    if channel == 'notice':
        CheckerEventHandler.check_notice(message)
    
    event_id, channel, message = EventServer.get_publish_message(channel, message)
    publish_queue.put((event_id, channel, message, args, kwargs))


def monkey_patch_issue(title, message, publish_queue, *args, **kwargs):
    from lyrebird.event import EventServer
    from lyrebird.checker.event import CheckerEventHandler
    notice = {
        "title": title,
        "message": message
    }
    CheckerEventHandler.check_notice(notice)

    event_id, channel, message = EventServer.get_publish_message('notice', notice)
    publish_queue.put((event_id, channel, message, args, kwargs))


class ProcessApplicationInfo(dict):

    def __init__(self, data=None, white_map={}):
        super().__init__()
        for path in white_map:
            value = self._get_value_from_path(data, path)
            if value is not None:
                self._set_value_to_path(path, value)

    def _get_value_from_path(self, data, path):
        keys = path.split('.')
        value = data
        for key in keys:
            value = getattr(value, key)
            if value is None:
                return None
        return value

    def _set_value_to_path(self, path, value):
        keys = path.split('.')
        current_dict = self
        for key in keys[:-1]:
            if key not in current_dict:
                current_dict[key] = ProcessApplicationInfo()
            current_dict = current_dict[key]
        current_dict[keys[-1]] = value

    
    def __getattr__(self, item):
        value = self
        for key in item.split('.'):
            value = value.get(key)
            if value is None:
                break
        return value

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, state):
        self.__dict__.update(state)

    def __setattr__(self, key, value):
        self[key] = value

    def __delattr__(self, item):
        del self[item]


compat_python_version_check()
compat_async_manager_to_python_3_8()



import os
import inspect
import traceback
from pathlib import Path
from collections import namedtuple

import pkg_resources
import setuptools
import importlib


from ..log import get_logger
from .plugin import Plugin

from lyrebird import application
import sys

logger = get_logger()


PLUGIN_ENTRY_POINT = "lyrebird_plugin"
manifest_cache = []


class ManifestError(Exception):
    pass


def manifest(**options):
    """
    Plugin entry method

        Usage:

        manifest(
            # Set UI entry point.
            # Each page need a tuple : (path, static-file-dir, html-file-name)
            # path - base url for plugin main page， userally it is '/'
            # static-file-dir - your frontent output dir
            # html-file-name (optional)  - default=index.html

            view=('/', 'static'),

            # Set plugin API.
            # Each API option is a tuple : (path, api-function, methods)
            # path - API path. If path='/status', the url will be 'http://lyrebid-host:port/plugin/plugin-name/api/status
            # api-function - API callback function. You can use flask.request for getting request info in this function.
            # methods - Same as flask methods.

            api=[
                ('status', plugin.mycallback, ['GET', 'POST'])
            ],

            # Set background task into a lyrebird task thread.

            background=[
                (task_name, task_function)
            ],

            # Set event listener
            # Each listener is a tuple : (channel, callback_function)
            # channel - Register listener to this channel
            # callback_function - This function will receive new event from the channel registered

            event=[
                ('flow', plugin.event_handler)
            ],

            # Set status bar text

            status=[
                plugin.StatusItem
            ]
        )
    """
    caller_info = inspect.stack()[1]
    manifest_cache.append((options, caller_info))


def is_plugin_enable(manifest_id):
    plugin_list = application.config.get('extension.plugin.enable')
    if plugin_list is None:
        return True
    return manifest_id in plugin_list


def load_all_from_ep():
    plugins = {}
    # Load plugin from installed packages by entry point 'lyrebird_plugin'
    for ep in pkg_resources.iter_entry_points(PLUGIN_ENTRY_POINT):
        try:
            plugin = load_plugin_from_ep(ep)
            if not is_plugin_enable(plugin.project_name):
                logger.info(f'Load plugin info: {plugin.project_name} is not enable.')
                continue

            if plugin.project_name in plugins:
                logger.error('Load plugin failed: More than one manifest in this plugin')
                continue
        except Exception as e:
            logger.error(f'Load plugin {ep.name} failed: {ep.dist}\n{e}\n{traceback.format_exc()}')
            continue
        plugins[plugin.project_name] = plugin
        logger.info(f'Load plugin from ep success: {plugin.project_name}')
    return plugins


def load_plugin_from_ep(ep):
    global manifest_cache
    manifest_cache = []

    ep_instance = ep.load()

    # There can only be one manifest in each plugin
    if len(manifest_cache) > 1:
        raise ManifestError('More than one manifest in this plugin')
    if len(manifest_cache) == 0:
        raise ManifestError('Not found manifest in plugin')

    # TODO
    manifest = manifest_cache[0][0]
    plugin = Plugin(
        manifest['id'],
        entry_point_name=ep.name,
        version=ep.dist.version,
        location=str(Path(ep_instance.__file__).parent),
        **manifest)

    return plugin


def load_from_path(plugin_path):
    global manifest_cache
    manifest_cache = []

    packages = setuptools.find_packages(plugin_path)
    for pkg in packages:
        manifest_file = Path(plugin_path)/pkg/'manifest.py'
        if not manifest_file.exists():
            continue
        if pkg in sys.modules:
            sys.modules.pop(pkg)
        if pkg+'.manifest' in sys.modules:
            sys.modules.pop(pkg+'.manifest')

        spec = importlib.util.spec_from_file_location(pkg, str(Path(plugin_path)/pkg/'__init__.py'))
        module = importlib.util.module_from_spec(spec)
        sys.modules[pkg] = module
        spec.loader.exec_module(module)

        __import__(pkg+'.manifest')

    # There can only one manifest in each plugin
    if len(manifest_cache) > 1:
        raise ManifestError('More than one manifest in this plugin')
    if len(manifest_cache) == 0:
        raise ManifestError(f'Not found any manifest in {plugin_path}')

    # TODO
    manifest = manifest_cache[0][0]
    caller_info = manifest_cache[0][1]
    plugin = Plugin(manifest['id'], **manifest)
    plugin.location = str(Path(caller_info.filename).parent)
    return plugin


def get_plugin_storage():
    """
    Get plugins storage dir path

    :return: ~/.lyrebird/plugins/<plugin_name>
    """
    info = _caller_info(index=2)
    storage_name = info.top_module_name
    application_conf_dir = os.path.join(os.path.expanduser('~'), '.lyrebird')
    plugin_storage_dir = os.path.abspath(os.path.join(application_conf_dir, 'plugins/%s' % storage_name))
    if not os.path.exists(plugin_storage_dir):
        os.makedirs(plugin_storage_dir)
    return plugin_storage_dir


def _caller_info(index=1):
    stack = inspect.stack()
    caller_module = inspect.getmodule(stack[index].frame)
    caller_module_name = caller_module.__name__
    if caller_module_name.find('.') > 0:
        caller_top_module_name = caller_module_name.split('.')[0]
    else:
        caller_top_module_name = caller_module_name
    CallerInfo = namedtuple('CallerInfo', 'top_module_name module_name')
    return CallerInfo(module_name=caller_module_name, top_module_name=caller_top_module_name)



CONFIG_TREE_SHOW_CONFIG = 'mock.data.shownConfig'
CONFIG_TREE_LOAD_CHILDREN_ASYNC= 'mock.data.tree.asynchronous'
CONFIG_MOCK_REQUEST_SSR_MOCK_IN_BODY = 'mock.request.ssr.mock_in_body'
CONFIG_MOCK_REQUEST_SSR_MOCK_BODY_KEY = 'mock.request.ssr.mock_body_key'
CONFIG_MOCK_REQUEST_SSR_MOCK_IN_BODY_PARAM = 'mock.request.ssr.mock_in_body.param'
CONFIG_MOCK_REQUEST_SSR_MOCK_DATA_JSONPATH = 'mock.request.ssr.mock_data_jsonpath'




import runpy
import os
import sys
from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))

VERSION = runpy.run_path(
    os.path.join(here, "lyrebird", "version.py")
)["VERSION"]


# Formatting (Spaces, etc.) in requirement must be consistent based on string parsing
def read_requirements(file_path):
    with open(file_path, encoding='utf-8') as f:
        return [
            line.strip().split(';')[0].strip()
            for line in f
            if line.strip() and not line.startswith('#') and (
                '; python_version' not in line or
                check_version_condition(line.split(';')[1].strip())
            )
        ]


def check_version_condition(condition):
    if not condition.startswith('python_version'):
        return True
    op, ver = condition.split(' ', 1)[1].split(' ')
    current_ver = sys.version_info[:2]
    ver = tuple(map(int, ver.strip('"').split('.')))
    return {
        '>': current_ver > ver,
        '>=': current_ver >= ver,
        '<': current_ver < ver,
        '<=': current_ver <= ver,
        '==': current_ver == ver
    }.get(op)


with open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name='lyrebird',
    version=VERSION,
    packages=['lyrebird'],
    url='https://github.com/meituan/lyrebird',
    author='HBQA',
    long_description=long_description,
    long_description_content_type="text/markdown",
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "License :: OSI Approved :: MIT License",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
    ],
    entry_points={
        'console_scripts': [
            'lyrebird = lyrebird.manager:main'
        ]
    },
    setup_requires=['packaging'],
    install_requires=read_requirements('requirements.txt.lock'),
    extras_require={
        'dev': read_requirements('requirements.dev.txt')
    }
)



import re
import platform
from packaging import version
from lyrebird.log import get_logger

logger = get_logger()


class JSONPath:

    @staticmethod
    def search(root, path:str, find_one=False):
        """ Find JSON object in object (a ``list`` or ``dict`` JSON object) file by JSONPath

        ``root`` <list> or <dict> an object containing data

        ``path`` <str> describe JSONPath of the target data, and must follow rules:
        1. Start with `$`
        2. Fuzzy search is not supported

        """

        if not path or not isinstance(path, str) or not isinstance(root, (list, dict)):
            return

        if path.startswith('$'):
            path = path.replace('$', '', 1)

        # split by `.` and drop `.`, split by `[ ]` and keep `[ ]`
        # EXAMPLE
        # path = 'data[*][1].name'
        # keys = ['data', '[*]', '[1]', 'name']

        pattern = r'(?:\.)|(?=\[.*\])'
        origin_keys = re.split(pattern, path)

        # There is a bug in re.split splitting with (?=) in Python 3.6 and below
        # The following code `re_split_handle` is used to solve this problem
        # Remove when Python 3.6 is not supported
        keys = [k for k in JSONPath.re_split_handle(origin_keys) if k]

        if not keys or not len(keys):
            return

        result = []
        JSONPath._search_iterator(root, keys, result)
        if find_one:
            return result[0].node if result else None
        return result

    @staticmethod
    def re_split_handle(origin_keys):
        # EXAMPLE
        # origin_keys = ['data[*][1]', 'name']
        # return keys = ['data', '[*]', '[1]', 'name']
        if version.parse(platform.python_version()) >= version.parse('3.7.0'):
            return origin_keys

        keys = []
        for key in origin_keys:
            keys.extend(re.findall('\[?[\w\*]+\]?', key))
        return keys

    @staticmethod
    def _search_iterator(root, prop_keys, result):
        current_key = prop_keys[0]

        keys = JSONPath.get_target_keys(root, current_key)
        for key in keys:
            if len(prop_keys) == 1:
                result.append(Node(root, key))
            else:
                JSONPath._search_iterator(root[key], prop_keys[1:], result)

    @staticmethod
    def get_target_keys(root, key):
        # EXAMPLE
        # matched [0], [10], [*]
        is_list = re.match('\[(\d+|\*)\]', key)

        if not is_list:
            if isinstance(root, dict) and (key in root):
                return (key,)
            else:
                return ()

        index = key.strip('[').strip(']')
        if not isinstance(root, list):
            return ()
        if index == '*':
            return range(0, len(root))
        if index.isdigit() and int(index) < len(root):
            return (int(index),)

        return ()


class Node:
    def __init__(self, parent, key):
        self.parent = parent
        self.node = parent[key]
        self.key = key

jsonpath = JSONPath()
